#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unit tests for IS_URL()
"""

import unittest

from pydal.validators import IS_URL, IS_HTTP_URL, IS_GENERIC_URL
from pydal.validators import unicode_to_ascii_authority


class TestIsUrl(unittest.TestCase):

    def testModeHttp(self):

        # defaults to mode='http'

        x = IS_URL()
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('http://google.ca', None))
        self.assertEqual(x('google.ca:80'), ('http://google.ca:80',
                         None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))
        self.assertEqual(x('google..ca'), ('google..ca', 'Enter a valid URL'))
        self.assertEqual(
            x('google.ca..'), ('google.ca..', 'Enter a valid URL'))

        # explicit use of 'http' mode

        x = IS_URL(mode='http')
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('http://google.ca', None))
        self.assertEqual(x('google.ca:80'), ('http://google.ca:80',
                         None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # prepends 'https' instead of 'http'

        x = IS_URL(mode='http', prepend_scheme='https')
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('https://google.ca', None))
        self.assertEqual(x('google.ca:80'), ('https://google.ca:80',
                         None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # prepending disabled

        x = IS_URL(prepend_scheme=None)
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('google.ca', None))
        self.assertEqual(x('google.ca:80'), ('google.ca:80', None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # custom allowed_schemes

        x = IS_URL(mode='http', allowed_schemes=[None, 'http'])
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('https://google.ca'), ('https://google.ca',
                         'Enter a valid URL'))
        self.assertEqual(x('google.ca'), ('http://google.ca', None))
        self.assertEqual(x('google.ca:80'), ('http://google.ca:80',
                         None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # custom allowed_schemes, excluding None

        x = IS_URL(allowed_schemes=['http'])
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('https://google.ca'), ('https://google.ca',
                         'Enter a valid URL'))
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))
        self.assertEqual(x('google.ca:80'), ('google.ca:80',
                         'Enter a valid URL'))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # custom allowed_schemes and prepend_scheme

        x = IS_URL(allowed_schemes=[None, 'https'],
                   prepend_scheme='https')
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         'Enter a valid URL'))
        self.assertEqual(x('https://google.ca'), ('https://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('https://google.ca', None))
        self.assertEqual(x('google.ca:80'), ('https://google.ca:80',
                         None))
        self.assertEqual(x('unreal.blargg'), ('unreal.blargg',
                         'Enter a valid URL'))

        # Now any URL requiring prepending will fail, but prepending is still
        # enabled!

        x = IS_URL(allowed_schemes=['http'])
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))

    def testModeGeneric(self):

        # 'generic' mode

        x = IS_URL(mode='generic')
        self.assertEqual(x('http://google.ca'), ('http://google.ca', None))
        self.assertEqual(x('google.ca'), ('google.ca', None))
        self.assertEqual(x('google.ca:80'), ('http://google.ca:80', None))
        self.assertEqual(x('blargg://unreal'), ('blargg://unreal',
                         'Enter a valid URL'))

        # 'generic' mode with custom allowed_schemes that still includes
        # 'http' (the default for prepend_scheme)

        x = IS_URL(mode='generic', allowed_schemes=['http', 'blargg'])
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('ftp://google.ca'), ('ftp://google.ca',
                         'Enter a valid URL'))
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))
        self.assertEqual(x('google.ca:80'), ('google.ca:80',
                         'Enter a valid URL'))
        self.assertEqual(x('blargg://unreal'), ('blargg://unreal',
                         None))

        # 'generic' mode with overriden prepend_scheme

        x = IS_URL(mode='generic', prepend_scheme='ftp')
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         None))
        self.assertEqual(x('ftp://google.ca'), ('ftp://google.ca',
                         None))
        self.assertEqual(x('google.ca'), ('google.ca', None))
        self.assertEqual(x('google.ca:80'), ('ftp://google.ca:80',
                         None))
        self.assertEqual(x('blargg://unreal'), ('blargg://unreal',
                         'Enter a valid URL'))

        # 'generic' mode with overriden allowed_schemes and prepend_scheme

        x = IS_URL(mode='generic', allowed_schemes=[None, 'ftp', 'ftps'
                                                    ], prepend_scheme='ftp')
        self.assertEqual(x('http://google.ca'), ('http://google.ca',
                         'Enter a valid URL'))
        self.assertEqual(x('google.ca'), ('google.ca', None))
        self.assertEqual(x('ftp://google.ca'), ('ftp://google.ca',
                         None))
        self.assertEqual(x('google.ca:80'), ('ftp://google.ca:80',
                         None))
        self.assertEqual(x('blargg://unreal'), ('blargg://unreal',
                         'Enter a valid URL'))

        # Now any URL requiring prepending will fail, but prepending is still
        # enabled!

        x = IS_URL(mode='generic', allowed_schemes=['http'])
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))

    def testExceptionalUse(self):

        # mode must be in set ['http', 'generic']

        try:
            x = IS_URL(mode='ftp')
            x('http://www.google.ca')
        except Exception as e:
            if str(e) != "invalid mode 'ftp' in IS_URL":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid mode: 'ftp'")

        # allowed_schemes in 'http' mode must be in set [None, 'http', 'https']

        try:
            x = IS_URL(allowed_schemes=[None, 'ftp', 'ftps'],
                       prepend_scheme='ftp')
            x('http://www.benn.ca')  # we can only reasonably know about the
                                     # error at calling time
        except Exception as e:
            if str(e)\
                    != "allowed_scheme value 'ftp' is not in [None, 'http', 'https']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail(
                "Accepted invalid allowed_schemes: [None, 'ftp', 'ftps']")

        # prepend_scheme's value must be in allowed_schemes (default for 'http'
        # mode is [None, 'http', 'https'])

        try:
            x = IS_URL(prepend_scheme='ftp')
            x('http://www.benn.ca')  # we can only reasonably know about the
                                     # error at calling time
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='ftp' is not in allowed_schemes=[None, 'http', 'https']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'ftp'")

        # custom allowed_schemes that excludes 'http', so prepend_scheme must be
        # specified!

        try:
            x = IS_URL(allowed_schemes=[None, 'https'])
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='http' is not in allowed_schemes=[None, 'https']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'http'")

        # prepend_scheme must be in allowed_schemes

        try:
            x = IS_URL(allowed_schemes=[None, 'http'],
                       prepend_scheme='https')
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='https' is not in allowed_schemes=[None, 'http']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'https'")

        # prepend_scheme's value (default is 'http') must be in allowed_schemes

        try:
            x = IS_URL(mode='generic', allowed_schemes=[None, 'ftp',
                       'ftps'])
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='http' is not in allowed_schemes=[None, 'ftp', 'ftps']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'http'")

        # prepend_scheme's value must be in allowed_schemes, which by default
        # is all schemes that really exist

        try:
            x = IS_URL(mode='generic', prepend_scheme='blargg')
            x('http://www.google.ca')
              # we can only reasonably know about the error at calling time
        except Exception as e:
            if not str(e).startswith(
                    "prepend_scheme='blargg' is not in allowed_schemes="):
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'blargg'")

        # prepend_scheme's value must be in allowed_schemes

        try:
            x = IS_URL(mode='generic', allowed_schemes=[None, 'http'],
                       prepend_scheme='blargg')
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='blargg' is not in allowed_schemes=[None, 'http']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'blargg'")

        # Not inluding None in the allowed_schemes essentially disabled
        # prepending, so even though
        # prepend_scheme has the invalid value 'http', we don't care!

        x = IS_URL(allowed_schemes=['https'], prepend_scheme='https')
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))

        # Not inluding None in the allowed_schemes essentially disabled prepending, so even though
        # prepend_scheme has the invalid value 'http', we don't care!

        x = IS_URL(mode='generic', allowed_schemes=['https'],
                   prepend_scheme='https')
        self.assertEqual(x('google.ca'), ('google.ca', 'Enter a valid URL'))


# ##############################################################################


class TestIsGenericUrl(unittest.TestCase):

    x = IS_GENERIC_URL()

    def testInvalidUrls(self):
        urlsToCheckA = []
        for i in list(range(0, 32)) + [127]:

            # Control characters are disallowed in any part of a URL

            urlsToCheckA.append('http://www.benn' + chr(i) + '.ca')

        urlsToCheckB = [
            None,
            '',
            'http://www.no spaces allowed.com',
            'http://www.benn.ca/no spaces allowed/',
            'http://www.benn.ca/angle_<bracket/',
            'http://www.benn.ca/angle_>bracket/',
            'http://www.benn.ca/invalid%character',
            'http://www.benn.ca/illegal%%20use',
            'http://www.benn.ca/illegaluse%',
            'http://www.benn.ca/illegaluse%0',
            'http://www.benn.ca/illegaluse%x',
            'http://www.benn.ca/ill%egaluse%x',
            'http://www.benn.ca/double"quote/',
            'http://www.curly{brace.com',
            'http://www.benn.ca/curly}brace/',
            'http://www.benn.ca/or|symbol/',
            'http://www.benn.ca/back\\slash',
            'http://www.benn.ca/the^carat',
            'http://left[bracket.me',
            'http://www.benn.ca/right]bracket',
            'http://www.benn.ca/angle`quote',
            '-ttp://www.benn.ca',
            '+ttp://www.benn.ca',
            '.ttp://www.benn.ca',
            '9ttp://www.benn.ca',
            'ht;tp://www.benn.ca',
            'ht@tp://www.benn.ca',
            'ht&tp://www.benn.ca',
            'ht=tp://www.benn.ca',
            'ht$tp://www.benn.ca',
            'ht,tp://www.benn.ca',
            'ht:tp://www.benn.ca',
            'htp://invalid_scheme.com',
        ]

        failures = []

        for url in urlsToCheckA + urlsToCheckB:
            if self.x(url)[1] is None:
                failures.append('Incorrectly accepted: ' + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testValidUrls(self):
        urlsToCheck = [
            'ftp://ftp.is.co.za/rfc/rfc1808.txt',
            'gopher://spinaltap.micro.umn.edu/00/Weather/California/Los%20Angeles',
            'http://www.math.uio.no/faq/compression-faq/part1.html',
            'mailto:mduerst@ifi.unizh.ch',
            'news:comp.infosystems.www.servers.unix',
            'telnet://melvyl.ucop.edu/',
            'hTTp://www.benn.ca',
            '%66%74%70://ftp.is.co.za/rfc/rfc1808.txt',
            '%46%74%70://ftp.is.co.za/rfc/rfc1808.txt',
            '/faq/compression-faq/part1.html',
            'google.com',
            'www.google.com:8080',
            '128.127.123.250:8080',
            'blargg:ping',
            'http://www.benn.ca',
            'http://benn.ca',
            'http://amazon.com/books/',
            'https://amazon.com/movies',
            'rtsp://idontknowthisprotocol',
            'HTTP://allcaps.com',
            'http://localhost',
            'http://localhost#fragment',
            'http://localhost/hello',
            'http://localhost/hello?query=True',
            'http://localhost/hello/',
            'http://localhost:8080',
            'http://localhost:8080/',
            'http://localhost:8080/hello',
            'http://localhost:8080/hello/',
            'file:///C:/Documents%20and%20Settings/Jonathan/Desktop/view.py',
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is not None:
                failures.append('Incorrectly rejected: ' + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testPrepending(self):
        # Does not prepend scheme for abbreviated domains
        self.assertEqual(self.x('google.ca'), ('google.ca', None))

        # Does not prepend scheme for abbreviated domains
        self.assertEqual(self.x('google.ca:8080'), ('google.ca:8080', None))

        # Does not prepend when scheme already exists
        self.assertEqual(self.x('https://google.ca'),
                         ('https://google.ca', None))

        # Does not prepend if None type is not specified in allowed_scheme,
        # because a scheme is required

        y = IS_GENERIC_URL(allowed_schemes=['http', 'blargg'],
                           prepend_scheme='http')
        self.assertEqual(y('google.ca'), ('google.ca', 'Enter a valid URL'))


# ##############################################################################


class TestIsHttpUrl(unittest.TestCase):

    x = IS_HTTP_URL()

    def testInvalidUrls(self):
        urlsToCheck = [
            None,
            '',
            'http://invalid' + chr(2) + '.com',
            'htp://invalid_scheme.com',
            'blargg://invalid_scheme.com',
            'http://-123.com',
            'http://abcd-.ca',
            'http://-abc123-.me',
            'http://www.dom&ain.com/',
            'http://www.dom=ain.com/',
            'http://www.benn.ca&',
            'http://%62%65%6E%6E%2E%63%61/path',
            'http://.domain.com',
            'http://.domain.com./path',
            'http://domain..com',
            'http://domain...at..com',
            'http://domain.com..',
            'http://domain.com../path',
            'http://domain.3m',
            'http://domain.-3m',
            'http://domain.3m-',
            'http://domain.-3m-',
            'http://domain.co&m',
            'http://domain.m3456',
            'http://domain.m-3/path#fragment',
            'http://domain.m---k/path?query=value',
            'http://23.32..',
            'http://23..32.56.0',
            'http://38997.222.999',
            'http://23.32.56.99.',
            'http://.23.32.56.99',
            'http://.23.32.56.99.',
            'http://w127.123.0.256:8080',
            'http://23.32.56.99:abcd',
            'http://23.32.56.99:23cd',
            'http://google.com:cd22',
            'http://23.32:1300.56.99',
            'http://www.yahoo:1600.com',
            'path/segment/without/starting/slash',
            'http://www.math.uio.no;param=3',
            '://ABC.com:/%7esmith/home.html',
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is None:
                failures.append('Incorrectly accepted: ' + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testValidUrls(self):

        urlsToCheck = [
            'http://abc.com:80/~smith/home.html',
            'http://ABC.com/%7Esmith/home.html',
            'http://ABC.com:/%7esmith/home.html',
            'http://www.math.uio.no/faq/compression-faq/part1.html',
            '//google.ca/faq/compression-faq/part1.html',
            '//google.ca/faq;param=3',
            '//google.ca/faq/index.html?query=5',
            '//google.ca/faq/index.html;param=value?query=5',
            '/faq/compression-faq/part1.html',
            '/faq;param=3',
            '/faq/index.html?query=5',
            '/faq/index.html;param=value?query=5',
            'google.com',
            'benn.ca/init/default',
            'benn.ca/init;param=value/default?query=value',
            'http://host-name---with-dashes.me',
            'http://www.host-name---with-dashes.me',
            'http://a.com',
            'http://a.3.com',
            'http://a.bl-ck.com',
            'http://bl-e.b.com',
            'http://host123with456numbers.ca',
            'http://1234567890.com.',
            'http://1234567890.com./path',
            'http://google.com./path',
            'http://domain.xn--d1acj3b',
            'http://127.123.0.256',
            'http://127.123.0.256/document/drawer',
            '127.123.0.256/document/',
            '156.212.123.100',
            'http://www.google.com:180200',
            'http://www.google.com:8080/path',
            'http://www.google.com:8080',
            '//www.google.com:8080',
            'www.google.com:8080',
            'http://127.123.0.256:8080/path',
            '//127.123.0.256:8080',
            '127.123.0.256:8080',
            'http://example.me??query=value?',
            'http://a.com',
            'http://3.com',
            'http://www.benn.ca',
            'http://benn.ca',
            'http://amazon.com/books/',
            'https://amazon.com/movies',
            'hTTp://allcaps.com',
            'http://localhost',
            'HTTPS://localhost.',
            'http://localhost#fragment',
            'http://localhost/hello;param=value',
            'http://localhost/hello;param=value/hi;param2=value2;param3=value3',
            'http://localhost/hello?query=True',
            'http://www.benn.ca/hello;param=value/hi;param2=value2;param3=value3/index.html?query=3',
            'http://localhost/hello/?query=1500&five=6',
            'http://localhost:8080',
            'http://localhost:8080/',
            'http://localhost:8080/hello',
            'http://localhost:8080/hello%20world/',
            'http://www.a.3.be-nn.5.ca',
            'http://www.amazon.COM',
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is not None:
                failures.append('Incorrectly rejected: ' + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testPrepending(self):
        # prepends scheme for abbreviated domains
        self.assertEqual(self.x('google.ca'), ('http://google.ca', None))

        # prepends scheme for abbreviated domains
        self.assertEqual(self.x('google.ca:8080'),
                         ('http://google.ca:8080', None))

        # does not prepend when scheme already exists
        self.assertEqual(self.x('https://google.ca'),
                         ('https://google.ca', None))

        y = IS_HTTP_URL(
            prepend_scheme='https', allowed_schemes=[None, 'https'])
        self.assertEqual(y('google.ca'), (
            'https://google.ca', None))  # prepends https if asked

        z = IS_HTTP_URL(prepend_scheme=None)
        self.assertEqual(z('google.ca:8080'), ('google.ca:8080', None))  # prepending disabled

        try:
            IS_HTTP_URL(prepend_scheme='mailto')
        except Exception as e:
            if str(e)\
                    != "prepend_scheme='mailto' is not in allowed_schemes=[None, 'http', 'https']":
                self.fail('Wrong exception: ' + str(e))
        else:
            self.fail("Got invalid prepend_scheme: 'mailto'")

        # Does not prepend if None type is not specified in allowed_scheme, because a scheme is required

        a = IS_HTTP_URL(allowed_schemes=['http'])
        self.assertEqual(a('google.ca'), ('google.ca', 'Enter a valid URL'))
        self.assertEqual(a('google.ca:80'), ('google.ca:80',
                         'Enter a valid URL'))


class TestUnicode(unittest.TestCase):
    x = IS_URL()
    y = IS_URL(allowed_schemes=['https'], prepend_scheme='https')
               #excludes the option for abbreviated URLs with no scheme
    z = IS_URL(prepend_scheme=None)
               # disables prepending the scheme in the return value

    def testUnicodeToAsciiUrl(self):
        self.assertEqual(unicode_to_ascii_authority(u'www.Alliancefran\xe7aise.nu'), 'www.xn--alliancefranaise-npb.nu')
        self.assertEqual(
            unicode_to_ascii_authority(u'www.benn.ca'), 'www.benn.ca')
        self.assertRaises(UnicodeError, unicode_to_ascii_authority,
                          u'\u4e2d' * 1000)  # label is too long

    def testValidUrls(self):
        self.assertEqual(self.x(u'www.Alliancefrancaise.nu'), (
            'http://www.Alliancefrancaise.nu', None))
        self.assertEqual(self.x(u'www.Alliancefran\xe7aise.nu'), (
            'http://www.xn--alliancefranaise-npb.nu', None))
        self.assertEqual(self.x(u'www.Alliancefran\xe7aise.nu:8080'), (
            'http://www.xn--alliancefranaise-npb.nu:8080', None))
        self.assertEqual(self.x(u'http://www.Alliancefran\xe7aise.nu'),
                          ('http://www.xn--alliancefranaise-npb.nu', None))
        self.assertEqual(self.x(u'http://www.Alliancefran\xe7aise.nu/parnaise/blue'), ('http://www.xn--alliancefranaise-npb.nu/parnaise/blue', None))
        self.assertEqual(self.x(u'http://www.Alliancefran\xe7aise.nu/parnaise/blue#fragment'), ('http://www.xn--alliancefranaise-npb.nu/parnaise/blue#fragment', None))
        self.assertEqual(self.x(u'http://www.Alliancefran\xe7aise.nu/parnaise/blue?query=value#fragment'), ('http://www.xn--alliancefranaise-npb.nu/parnaise/blue?query=value#fragment', None))
        self.assertEqual(self.x(u'http://www.Alliancefran\xe7aise.nu:8080/parnaise/blue?query=value#fragment'), ('http://www.xn--alliancefranaise-npb.nu:8080/parnaise/blue?query=value#fragment', None))
        self.assertEqual(self.x(u'www.Alliancefran\xe7aise.nu/parnaise/blue?query=value#fragment'), ('http://www.xn--alliancefranaise-npb.nu/parnaise/blue?query=value#fragment', None))
        self.assertEqual(self.x(
            u'http://\u4e2d\u4fd4.com'), ('http://xn--fiq13b.com', None))
        self.assertEqual(self.x(u'http://\u4e2d\u4fd4.com/\u4e86'),
                          ('http://xn--fiq13b.com/%4e%86', None))
        self.assertEqual(self.x(u'http://\u4e2d\u4fd4.com/\u4e86?query=\u4e86'), ('http://xn--fiq13b.com/%4e%86?query=%4e%86', None))
        self.assertEqual(self.x(u'http://\u4e2d\u4fd4.com/\u4e86?query=\u4e86#fragment'), ('http://xn--fiq13b.com/%4e%86?query=%4e%86#fragment', None))
        self.assertEqual(self.x(u'http://\u4e2d\u4fd4.com?query=\u4e86#fragment'), ('http://xn--fiq13b.com?query=%4e%86#fragment', None))
        self.assertEqual(
            self.x(u'http://B\xfccher.ch'), ('http://xn--bcher-kva.ch', None))
        self.assertEqual(self.x(u'http://\xe4\xf6\xfc\xdf.com'), (
            'http://xn--ss-uia6e4a.com', None))
        self.assertEqual(self.x(
            u'http://visegr\xe1d.com'), ('http://xn--visegrd-mwa.com', None))
        self.assertEqual(self.x(u'http://h\xe1zipatika.com'), (
            'http://xn--hzipatika-01a.com', None))
        self.assertEqual(self.x(u'http://www.\xe7ukurova.com'), (
            'http://www.xn--ukurova-txa.com', None))
        self.assertEqual(self.x(u'http://nixier\xf6hre.nixieclock-tube.com'), ('http://xn--nixierhre-57a.nixieclock-tube.com', None))
        self.assertEqual(self.x(u'google.ca.'), ('http://google.ca.', None))

        self.assertEqual(
            self.y(u'https://google.ca'), ('https://google.ca', None))
        self.assertEqual(self.y(
            u'https://\u4e2d\u4fd4.com'), ('https://xn--fiq13b.com', None))

        self.assertEqual(self.z(u'google.ca'), ('google.ca', None))

    def testInvalidUrls(self):
        self.assertEqual(
            self.x(u'://ABC.com'), (u'://ABC.com', 'Enter a valid URL'))
        self.assertEqual(self.x(u'http://\u4e2d\u4fd4.dne'), (
            u'http://\u4e2d\u4fd4.dne', 'Enter a valid URL'))
        self.assertEqual(self.x(u'https://google.dne'), (
            u'https://google.dne', 'Enter a valid URL'))
        self.assertEqual(self.x(u'https://google..ca'), (
            u'https://google..ca', 'Enter a valid URL'))
        self.assertEqual(
            self.x(u'google..ca'), (u'google..ca', 'Enter a valid URL'))
        self.assertEqual(self.x(u'http://' + u'\u4e2d' * 1000 + u'.com'), (
            u'http://' + u'\u4e2d' * 1000 + u'.com', 'Enter a valid URL'))

        self.assertEqual(self.x(u'http://google.com#fragment_\u4e86'), (
            u'http://google.com#fragment_\u4e86', 'Enter a valid URL'))
        self.assertEqual(self.x(u'http\u4e86://google.com'), (
            u'http\u4e86://google.com', 'Enter a valid URL'))
        self.assertEqual(self.x(u'http\u4e86://google.com#fragment_\u4e86'), (
            u'http\u4e86://google.com#fragment_\u4e86', 'Enter a valid URL'))

        self.assertEqual(self.y(u'http://\u4e2d\u4fd4.com/\u4e86'), (
            u'http://\u4e2d\u4fd4.com/\u4e86', 'Enter a valid URL'))
        #self.assertEqual(self.y(u'google.ca'), (u'google.ca', 'Enter a valid URL'))

        self.assertEqual(self.z(u'invalid.domain..com'), (
            u'invalid.domain..com', 'Enter a valid URL'))
        self.assertEqual(self.z(u'invalid.\u4e2d\u4fd4.blargg'), (
            u'invalid.\u4e2d\u4fd4.blargg', 'Enter a valid URL'))

# ##############################################################################


class TestSimple(unittest.TestCase):

    def test_IS_URL(self):
        rtn = IS_URL()('abc.com')
        self.assertEqual(rtn, ('http://abc.com', None))
        rtn = IS_URL(mode='generic')('abc.com')
        self.assertEqual(rtn, ('abc.com', None))
        rtn = IS_URL(allowed_schemes=['https'], prepend_scheme='https')('https://abc.com')
        self.assertEqual(rtn, ('https://abc.com', None))
        rtn = IS_URL(prepend_scheme='https')('abc.com')
        self.assertEqual(rtn, ('https://abc.com', None))
        rtn = IS_URL(mode='generic', allowed_schemes=['ftps', 'https'], prepend_scheme='https')('https://abc.com')
        self.assertEqual(rtn, ('https://abc.com', None))
        rtn = IS_URL(mode='generic', allowed_schemes=['ftps', 'https', None], prepend_scheme='https')('abc.com')
        self.assertEqual(rtn, ('abc.com', None))
        # regression test for issue 773
        rtn = IS_URL()('domain.ninja')
        self.assertEqual(rtn, ('http://domain.ninja', None))
        # addition of allowed_tlds
        rtn = IS_URL(allowed_tlds=['com', 'net', 'org'])('domain.ninja')
        self.assertEqual(rtn, ('domain.ninja', 'Enter a valid URL'))
        # mode = 'generic' doesn't consider allowed_tlds
        rtn = IS_URL(mode='generic', allowed_tlds=['com', 'net', 'org'])('domain.ninja')
        self.assertEqual(rtn, ('domain.ninja', None))



from .._compat import basestring
from ..adapters.teradata import Teradata
from .base import SQLDialect
from . import dialects, sqltype_for


@dialects.register_for(Teradata)
class TeradataDialect(SQLDialect):
    @sqltype_for('integer')
    def type_integer(self):
        return 'INT'

    @sqltype_for('text')
    def type_text(self):
        return 'VARCHAR(2000)'

    @sqltype_for('json')
    def type_json(self):
        return 'VARCHAR(4000)'

    @sqltype_for('float')
    def type_float(self):
        return 'REAL'

    @sqltype_for('list:integer')
    def type_list_integer(self):
        return self.types['json']

    @sqltype_for('list:string')
    def type_list_string(self):
        return self.types['json']

    @sqltype_for('list:reference')
    def type_list_reference(self):
        return self.types['json']

    @sqltype_for('bigint')
    def type_bigint(self):
        return 'BIGINT'

    @sqltype_for('id')
    def type_id(self):
        return 'INT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL'

    @sqltype_for('big-id')
    def type_big_id(self):
        return 'BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL'

    @sqltype_for('reference')
    def type_reference(self):
        return 'INT'

    @sqltype_for('big-reference')
    def type_big_reference(self):
        return 'BIGINT'

    @sqltype_for('geometry')
    def type_geometry(self):
        return 'ST_GEOMETRY'

    @sqltype_for('reference FK')
    def type_reference_fk(self):
        return ' REFERENCES %(foreign_key)s '

    @sqltype_for('reference TFK')
    def type_reference_tfk(self):
        return ' FOREIGN KEY (%(field_name)s) REFERENCES %(foreign_table)s' + \
            ' (%(foreign_key)s)'

    def left_join(self, val, query_env={}):
        # Left join must always have an ON clause
        if not isinstance(val, basestring):
            val = self.expand(val, query_env=query_env)
        return 'LEFT OUTER JOIN %s' % val

    def select(self, fields, tables, where=None, groupby=None, having=None,
               orderby=None, limitby=None, distinct=False, for_update=False):
        dst, whr, grp, order, limit, offset, upd = '', '', '', '', '', '', ''
        if distinct is True:
            dst = ' DISTINCT'
        elif distinct:
            dst = ' DISTINCT ON (%s)' % distinct
        if where:
            whr = ' %s' % self.where(where)
        if groupby:
            grp = ' GROUP BY %s' % groupby
            if having:
                grp += ' HAVING %s' % having
        if orderby:
            order = ' ORDER BY %s' % orderby
        if limitby:
            (lmin, lmax) = limitby
            limit = ' TOP %i' % lmax
        if for_update:
            upd = ' FOR UPDATE'
        return 'SELECT%s%s %s FROM %s%s%s%s%s%s;' % (
            dst, limit, fields, tables, whr, grp, order, offset, upd)

    def truncate(self, table, mode=''):
        return ['DELETE FROM %s ALL;' % table._rname]



import re
from .regex import REGEX_SEARCH_PATTERN, REGEX_SQUARE_BRACKETS
from .._compat import long

def to_num(num):
    result = 0
    try:
        result = long(num)
    except NameError as e:
        result = int(num)
    return result

class RestParser(object):
    def __init__(self, db):
        self.db = db

    def auto_table(self, table, base='', depth=0):
        patterns = []
        for field in self.db[table].fields:
            if base:
                tag = '%s/%s' % (base, field.replace('_', '-'))
            else:
                tag = '/%s/%s' % (
                    table.replace('_', '-'), field.replace('_', '-'))
            f = self.db[table][field]
            if not f.readable:
                continue
            if f.type == 'id' or 'slug' in field or \
               f.type.startswith('reference'):
                tag += '/{%s.%s}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            elif f.type.startswith('boolean'):
                tag += '/{%s.%s}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            elif f.type in ('float', 'double', 'integer', 'bigint'):
                tag += '/{%s.%s.ge}/{%s.%s.lt}' % (table, field, table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            elif f.type.startswith('list:'):
                tag += '/{%s.%s.contains}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            elif f.type in ('date', 'datetime'):
                tag += '/{%s.%s.year}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
                tag += '/{%s.%s.month}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
                tag += '/{%s.%s.day}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            if f.type in ('datetime', 'time'):
                tag += '/{%s.%s.hour}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
                tag += '/{%s.%s.minute}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
                tag += '/{%s.%s.second}' % (table, field)
                patterns.append(tag)
                patterns.append(tag+'/:field')
            if depth > 0:
                for f in self.db[table]._referenced_by:
                    tag += '/%s[%s.%s]' % (table, f.tablename, f.name)
                    patterns.append(tag)
                    patterns += self.auto_table(table, base=tag, depth=depth-1)
        return patterns

    def parse(self, patterns, args, vars, queries=None, nested_select=True):
        """
        Example:
            Use as::

                db.define_table('person',Field('name'),Field('info'))
                db.define_table('pet',
                    Field('ownedby',db.person),
                    Field('name'),Field('info')
                )

                @request.restful()
                def index():
                    def GET(*args,**vars):
                        patterns = [
                            "/friends[person]",
                            "/{person.name}/:field",
                            "/{person.name}/pets[pet.ownedby]",
                            "/{person.name}/pets[pet.ownedby]/{pet.name}",
                            "/{person.name}/pets[pet.ownedby]/{pet.name}/:field",
                            ("/dogs[pet]", db.pet.info=='dog'),
                            ("/dogs[pet]/{pet.name.startswith}", db.pet.info=='dog'),
                            ]
                        parser = db.parse_as_rest(patterns,args,vars)
                        if parser.status == 200:
                            return dict(content=parser.response)
                        else:
                            raise HTTP(parser.status,parser.error)

                    def POST(table_name,**vars):
                        if table_name == 'person':
                            return db.person.validate_and_insert(**vars)
                        elif table_name == 'pet':
                            return db.pet.validate_and_insert(**vars)
                        else:
                            raise HTTP(400)
                    return locals()
        """

        if patterns == 'auto':
            patterns = []
            for table in self.db.tables:
                if not table.startswith('auth_'):
                    patterns.append('/%s[%s]' % (table, table))
                    patterns += self.auto_table(table, base='', depth=1)
        else:
            i = 0
            while i < len(patterns):
                pattern = patterns[i]
                if not isinstance(pattern, str):
                    pattern = pattern[0]
                tokens = pattern.split('/')
                if tokens[-1].startswith(':auto') and \
                        re.match(REGEX_SQUARE_BRACKETS, tokens[-1]):
                    new_patterns = self.auto_table(
                        tokens[-1][tokens[-1].find('[')+1:-1],
                        '/'.join(tokens[:-1]))
                    patterns = patterns[:i]+new_patterns+patterns[i+1:]
                    i += len(new_patterns)
                else:
                    i += 1
        if '/'.join(args) == 'patterns':
            return self.db.Row({
                'status': 200, 'pattern': 'list', 'error': None,
                'response': patterns})
        for pattern in patterns:
            basequery, exposedfields = None, []
            if isinstance(pattern, tuple):
                if len(pattern) == 2:
                    pattern, basequery = pattern
                elif len(pattern) > 2:
                    pattern, basequery, exposedfields = pattern[0:3]
            otable = table = None
            if not isinstance(queries, dict):
                dbset = self.db(queries)
                if basequery is not None:
                    dbset = dbset(basequery)
            i = 0
            tags = pattern[1:].split('/')
            if len(tags) != len(args):
                continue
            for tag in tags:
                if re.match(REGEX_SEARCH_PATTERN, tag):
                    tokens = tag[1:-1].split('.')
                    table, field = tokens[0], tokens[1]
                    if not otable or table == otable:
                        if len(tokens) == 2 or tokens[2] == 'eq':
                            query = self.db[table][field] == args[i]
                        elif tokens[2] == 'ne':
                            query = self.db[table][field] != args[i]
                        elif tokens[2] == 'lt':
                            query = self.db[table][field] < args[i]
                        elif tokens[2] == 'gt':
                            query = self.db[table][field] > args[i]
                        elif tokens[2] == 'ge':
                            query = self.db[table][field] >= args[i]
                        elif tokens[2] == 'le':
                            query = self.db[table][field] <= args[i]
                        elif tokens[2] == 'year':
                            query = self.db[table][field].year() == args[i]
                        elif tokens[2] == 'month':
                            query = self.db[table][field].month() == args[i]
                        elif tokens[2] == 'day':
                            query = self.db[table][field].day() == args[i]
                        elif tokens[2] == 'hour':
                            query = self.db[table][field].hour() == args[i]
                        elif tokens[2] == 'minute':
                            query = self.db[table][field].minutes() == args[i]
                        elif tokens[2] == 'second':
                            query = self.db[table][field].seconds() == args[i]
                        elif tokens[2] == 'startswith':
                            query = self.db[table][field].startswith(args[i])
                        elif tokens[2] == 'contains':
                            query = self.db[table][field].contains(args[i])
                        else:
                            raise RuntimeError("invalid pattern: %s" % pattern)
                        if len(tokens) == 4 and tokens[3] == 'not':
                            query = ~query
                        elif len(tokens) >= 4:
                            raise RuntimeError("invalid pattern: %s" % pattern)
                        if not otable and isinstance(queries, dict):
                            dbset = self.db(queries[table])
                            if basequery is not None:
                                dbset = dbset(basequery)
                        dbset = dbset(query)
                    else:
                        raise RuntimeError(
                            "missing relation in pattern: %s" % pattern)
                elif re.match(REGEX_SQUARE_BRACKETS, tag) and \
                        args[i] == tag[:tag.find('[')]:
                    ref = tag[tag.find('[')+1:-1]
                    if '.' in ref and otable:
                        table, field = ref.split('.')
                        selfld = '_id'
                        if self.db[table][field].type.startswith('reference '):
                            refs = [
                                x.name for x in self.db[otable]
                                if x.type == self.db[table][field].type]
                        else:
                            refs = [
                                x.name for x in self.db[table]._referenced_by
                                if x.tablename == otable]
                        if refs:
                            selfld = refs[0]
                        if nested_select:
                            try:
                                dbset = self.db(self.db[table][field].belongs(
                                    dbset._select(self.db[otable][selfld])))
                            except ValueError:
                                return self.db.Row({
                                    'status': 400, 'pattern': pattern,
                                    'error': 'invalid path', 'response': None})
                        else:
                            items = [
                                item.id for item in dbset.select(
                                    self.db[otable][selfld])
                            ]
                            dbset = self.db(
                                self.db[table][field].belongs(items))
                    else:
                        table = ref
                        if not otable and isinstance(queries, dict):
                            dbset = self.db(queries[table])
                        dbset = dbset(self.db[table])
                elif tag == ':field' and table:
                    # print 're3:'+tag
                    field = args[i]
                    if field not in self.db[table]:
                        break
                    # hand-built patterns should respect .readable=False as well
                    if not self.db[table][field].readable:
                        return self.db.Row({
                            'status': 418, 'pattern': pattern,
                            'error': 'I\'m a teapot', 'response': None})
                    try:
                        distinct = vars.get('distinct', False) == 'True'
                        offset = to_num(vars.get('offset', None) or 0)
                        limits = (
                            offset,
                            to_num(vars.get('limit', None) or 1000) + offset)
                    except ValueError:
                        return self.db.Row({
                            'status': 400, 'error': 'invalid limits',
                            'response': None})
                    items = dbset.select(
                        self.db[table][field], distinct=distinct,
                        limitby=limits)
                    if items:
                        return self.db.Row({
                            'status': 200, 'response': items,
                            'pattern': pattern})
                    else:
                        return self.db.Row({
                            'status': 404, 'pattern': pattern,
                            'error': 'no record found', ' response': None})
                elif tag != args[i]:
                    break
                otable = table
                i += 1
                if i == len(tags) and table:
                    if hasattr(self.db[table], '_id'):
                        ofields = vars.get(
                            'order', self.db[table]._id.name).split('|')
                    else:
                        ofields = vars.get(
                            'order', self.db[table]._primarykey[0]).split('|')
                    try:
                        orderby = [
                            self.db[table][f] if not f.startswith('~')
                            else ~self.db[table][f[1:]] for f in ofields]
                    except (KeyError, AttributeError):
                        return self.db.Row({
                            'status': 400, 'error': 'invalid orderby',
                            'response': None})
                    if exposedfields:
                        fields = [
                            field for field in self.db[table]
                            if str(field).split('.')[-1] in exposedfields and
                            field.readable]
                    else:
                        fields = [
                            field for field in self.db[table]
                            if field.readable]
                    count = dbset.count()
                    try:
                        offset = to_num(vars.get('offset', None) or 0)
                        limits = (
                            offset,
                            to_num(vars.get('limit', None) or 1000) + offset)
                    except ValueError:
                        return self.db.Row({
                            'status': 400, 'error': ' invalid limits',
                            'response': None})
                    try:
                        response = dbset.select(
                            limitby=limits, orderby=orderby, *fields)
                    except ValueError:
                        return self.db.Row({
                            'status': 400, 'pattern': pattern,
                            'error': 'invalid path', 'response': None})
                    return self.db.Row({
                        'status': 200, 'response': response,
                        'pattern': pattern, 'count': count})
        return self.db.Row({
            'status': 400, 'error': 'no matching pattern', 'response': None})



from ..adapters.sqlite import SQLite, Spatialite
from .base import SQLRepresenter, JSONRepresenter
from . import representers, before_type, for_type


@representers.register_for(SQLite)
class SQLiteRepresenter(SQLRepresenter, JSONRepresenter):
    pass


@representers.register_for(Spatialite)
class SpatialiteRepresenter(SQLRepresenter):
    @before_type('geometry')
    def geometry_extras(self, field_type):
        srid = 4326
        geotype, params = field_type[:-1].split('(')
        params = params.split(',')
        if len(params) >= 2:
            schema, srid = params[:2]
        return {'srid': srid}

    @for_type('geometry', adapt=False)
    def _geometry(self, value, srid):
        return "ST_GeomFromText('%s',%s)" % (value, srid)



from collections import defaultdict
from .._compat import PY2, with_metaclass, iteritems, to_unicode, to_bytes, \
    string_types
from .._gae import gae
from ..helpers._internals import Dispatcher
from ..helpers.regex import REGEX_TYPE

representers = Dispatcher("representer")


class for_type(object):
    def __init__(self, field_type, encode=False, adapt=True):
        self.field_type = field_type
        self.encode = encode
        self.adapt = adapt

    def __call__(self, f):
        self.f = f
        return self


class before_type(object):
    def __init__(self, field_type):
        self.field_type = field_type

    def __call__(self, f):
        self.f = f
        return self


class for_instance(object):
    def __init__(self, inst_type, repr_type=False):
        self.inst_type = inst_type
        self.repr_type = repr_type

    def __call__(self, f):
        self.f = f
        return self


class pre(object):
    _inst_count_ = 0

    def __init__(self, is_breaking=None):
        self.breaking = is_breaking
        self._inst_count_ = pre._inst_count_
        pre._inst_count_ += 1

    def __call__(self, f):
        self.f = f
        return self


class MetaRepresenter(type):
    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if bases == (object,):
            return new_class
        #: collect declared attributes
        trepresenters = {}
        irepresenters = {}
        tbefore = {}
        pres = {}
        for key, value in list(attrs.items()):
            if isinstance(value, for_type):
                trepresenters[key] = value
            elif isinstance(value, before_type):
                tbefore[key] = value
            elif isinstance(value, for_instance):
                irepresenters[key] = value
            elif isinstance(value, pre):
                pres[key] = value
        #: get super declared attributes
        declared_trepresenters = {}
        declared_irepresenters = {}
        declared_tbefore = {}
        declared_pres = {}
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, '_declared_trepresenters_'):
                declared_trepresenters.update(base._declared_trepresenters_)
            if hasattr(base, '_declared_irepresenters_'):
                declared_irepresenters.update(base._declared_irepresenters_)
            if hasattr(base, '_declared_tbefore_'):
                declared_tbefore.update(base._declared_tbefore_)
            if hasattr(base, '_declared_pres_'):
                declared_pres.update(base._declared_pres_)
        #: set trepresenters
        declared_trepresenters.update(trepresenters)
        declared_irepresenters.update(irepresenters)
        declared_tbefore.update(tbefore)
        declared_pres.update(pres)
        new_class._declared_trepresenters_ = declared_trepresenters
        new_class._declared_irepresenters_ = declared_irepresenters
        new_class._declared_tbefore_ = declared_tbefore
        new_class._declared_pres_ = declared_pres
        return new_class


class TReprMethodWrapper(object):
    def __init__(self, representer, obj, extra=None):
        self.representer = representer
        self.obj = obj
        if extra:
            self.extra = extra
            self.call = self._call_with_extras
        else:
            self.call = self._call
        if self.obj.encode and PY2:
            self.inner_call = self._inner_call_with_encode
        else:
            self.inner_call = self._inner_call
        if self.obj.adapt:
            self.adapt = self._adapt
        else:
            self.adapt = self._no_adapt

    def _adapt(self, value):
        return self.representer.adapt(value)

    def _no_adapt(self, value):
        return value

    def _inner_call(self, value, **kwargs):
        return self.obj.f(self.representer, value, **kwargs)

    def _inner_call_with_encode(self, value, **kwargs):
        if isinstance(value, unicode):
            value = value.encode(self.representer.adapter.db_codec)
        return self.obj.f(self.representer, value, **kwargs)

    def _call_with_extras(self, value, field_type):
        extras = self.extra(self.representer, field_type)
        return self.inner_call(value, **extras)

    def _call(self, value, field_type):
        return self.inner_call(value)

    def __call__(self, value, field_type):
        return self.adapt(self.call(value, field_type))


class IReprMethodWrapper(object):
    def __init__(self, representer, obj):
        self.representer = representer
        self.obj = obj

    def __call__(self, value, field_type):
        rv = self.obj.f(self.representer, value, field_type)
        return self.obj.repr_type, rv


class PreMethodWrapper(object):
    def __init__(self, representer, obj):
        self.representer = representer
        self.obj = obj
        if self.obj.breaking is None:
            self.call = self._call_autobreak
        elif self.obj.breaking == True:
            self.call = self._call_break
        else:
            self.call = self._call_nobreak

    def _call_autobreak(self, value, field_type):
        rv = self.obj.f(self.representer, value, field_type)
        if rv is not None:
            return True, rv
        return False, value

    def _call_break(self, value, field_type):
        return self.obj.f(
            self.representer, value, field_type)

    def _call_nobreak(self, value, field_type):
        return False, self.obj.f(self.representer, value, field_type)

    def __call__(self, value, field_type):
        return self.call(value, field_type)


class Representer(with_metaclass(MetaRepresenter)):
    def __init__(self, adapter):
        self.adapter = adapter
        self.dialect = adapter.dialect
        self._tbefore_registry_ = {}
        for name, obj in iteritems(self._declared_tbefore_):
            self._tbefore_registry_[obj.field_type] = obj.f
        self.registered_t = defaultdict(lambda self=self: self._default)
        for name, obj in iteritems(self._declared_trepresenters_):
            if obj.field_type in self._tbefore_registry_:
                self.registered_t[obj.field_type] = TReprMethodWrapper(
                    self, obj, self._tbefore_registry_[obj.field_type]
                )
            else:
                self.registered_t[obj.field_type] = TReprMethodWrapper(
                    self, obj
                )
        self.registered_i = {}
        for name, obj in iteritems(self._declared_irepresenters_):
            self.registered_i[obj.inst_type] = IReprMethodWrapper(self, obj)
        self._pre_registry_ = []
        pres = []
        for name, obj in iteritems(self._declared_pres_):
            pres.append(obj)
        pres.sort(key=lambda x: x._inst_count_)
        for pre in pres:
            self._pre_registry_.append(PreMethodWrapper(self, pre))

    def _default(self, value, field_type):
        return self.adapt(value)

    def _default_instance(self, value, field_type):
        return True, value

    def get_representer_for_instance(self, value):
        for inst, representer in iteritems(self.registered_i):
            if isinstance(value, inst):
                return representer
        return self._default_instance

    def get_representer_for_type(self, field_type):
        key = REGEX_TYPE.match(field_type).group(0)
        return self.registered_t[key]

    def adapt(self, value):
        if PY2:
            if not isinstance(value, string_types):
                value = str(value)
            value = to_bytes(value)
            try:
                value.decode(self.adapter.db_codec)
            except:
                value = value.decode('latin1').encode(self.adapter.db_codec)
        else:
            value = to_unicode(value)
        return self.adapter.adapt(value)

    def exceptions(self, value, field_type):
        return None

    def represent(self, value, field_type):
        pre_end = False
        for pre in self._pre_registry_:
            pre_end, value = pre(value, field_type)
            if pre_end:
                break
        if pre_end:
            return value
        repr_type, rv = self.get_representer_for_instance(value)(
            value, field_type)
        if repr_type:
            rv = self.get_representer_for_type(field_type)(rv, field_type)
        return rv


from .base import BaseRepresenter, SQLRepresenter, NoSQLRepresenter
from .sqlite import SQLiteRepresenter, SpatialiteRepresenter
from .postgre import PostgreRepresenter
from .mysql import MySQLRepresenter
from .mssql import MSSQLRepresenter
from .mongo import MongoRepresenter
from .db2 import DB2Representer
from .informix import InformixRepresenter
from .oracle import OracleRepresenter
from .couchdb import CouchDBRepresenter

if gae is not None:
    from .google import GoogleDatastoreRepresenter



from ..adapters.postgres import Postgre, PostgreNew, PostgreBoolean
from .._compat import integer_types, basestring
from ..helpers.methods import varquote_aux
from ..objects import Expression
from .base import SQLDialect
from . import dialects, sqltype_for, register_expression


@dialects.register_for(Postgre)
class PostgreDialect(SQLDialect):
    true_exp = "TRUE"
    false_exp = "FALSE"

    @sqltype_for('blob')
    def type_blob(self):
        return 'BYTEA'

    @sqltype_for('bigint')
    def type_bigint(self):
        return 'BIGINT'

    @sqltype_for('double')
    def type_double(self):
        return 'FLOAT8'

    @sqltype_for('id')
    def type_id(self):
        return 'SERIAL PRIMARY KEY'

    @sqltype_for('big-id')
    def type_big_id(self):
        return 'BIGSERIAL PRIMARY KEY'

    @sqltype_for('big-reference')
    def type_big_reference(self):
        return 'BIGINT REFERENCES %(foreign_key)s ' + \
            'ON DELETE %(on_delete_action)s %(null)s %(unique)s'

    @sqltype_for('reference TFK')
    def type_reference_tfk(self):
        return ' CONSTRAINT "FK_%(constraint_name)s_PK" FOREIGN KEY ' + \
            '(%(field_name)s) REFERENCES %(foreign_table)s' + \
            '(%(foreign_key)s) ON DELETE %(on_delete_action)s'

    @sqltype_for('geometry')
    def type_geometry(self):
        return 'GEOMETRY'

    @sqltype_for('geography')
    def type_geography(self):
        return 'GEOGRAPHY'

    def varquote(self, val):
        return varquote_aux(val, '"%s"')

    def sequence_name(self, tablename):
        return self.quote('%s_id_seq' % tablename)

    def insert(self, table, fields, values, returning=None):
        ret = ''
        if returning:
            ret = 'RETURNING %s' % returning
        return 'INSERT INTO %s(%s) VALUES (%s)%s;' % (
            table, fields, values, ret)

    @property
    def random(self):
        return 'RANDOM()'

    def add(self, first, second, query_env={}):
        t = first.type
        if t in ('text', 'string', 'password', 'json', 'jsonb', 'upload', 'blob'):
            return '(%s || %s)' % (
                self.expand(first, query_env=query_env),
                self.expand(second, first.type, query_env=query_env))
        else:
            return '(%s + %s)' % (
                self.expand(first, query_env=query_env),
                self.expand(second, first.type, query_env=query_env))

    def regexp(self, first, second, query_env={}):
        return '(%s ~ %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, 'string', query_env=query_env))

    def like(self, first, second, escape=None, query_env={}):
        if isinstance(second, Expression):
            second = self.expand(second, 'string', query_env=query_env)
        else:
            second = self.expand(second, 'string', query_env=query_env)
            if escape is None:
                escape = '\\'
                second = second.replace(escape, escape * 2)
        if first.type not in ('string', 'text', 'json', 'jsonb'):
            return "(%s LIKE %s ESCAPE '%s')" % (
                self.cast(self.expand(first, query_env=query_env),
                'CHAR(%s)' % first.length), second, escape)
        return "(%s LIKE %s ESCAPE '%s')" % (
            self.expand(first, query_env=query_env), second, escape)

    def ilike(self, first, second, escape=None, query_env={}):
        if isinstance(second, Expression):
            second = self.expand(second, 'string', query_env=query_env)
        else:
            second = self.expand(second, 'string', query_env=query_env)
            if escape is None:
                escape = '\\'
                second = second.replace(escape, escape * 2)
        if first.type not in ('string', 'text', 'json', 'jsonb', 'list:string'):
            return "(%s ILIKE %s ESCAPE '%s')" % (
                self.cast(self.expand(first, query_env=query_env),
                'CHAR(%s)' % first.length), second, escape)
        return "(%s ILIKE %s ESCAPE '%s')" % (
            self.expand(first, query_env=query_env), second, escape)

    def drop_table(self, table, mode):
        if mode not in ['restrict', 'cascade', '']:
            raise ValueError('Invalid mode: %s' % mode)
        return ['DROP TABLE ' + table._rname + ' ' + mode + ';']

    def create_index(self, name, table, expressions, unique=False, where=None):
        uniq = ' UNIQUE' if unique else ''
        whr = ''
        if where:
            whr = ' %s' % self.where(where)
        with self.adapter.index_expander():
            rv = 'CREATE%s INDEX %s ON %s (%s)%s;' % (
                uniq, self.quote(name), table._rname, ','.join(
                    self.expand(field) for field in expressions), whr)
        return rv

    def st_asgeojson(self, first, second, query_env={}):
        return 'ST_AsGeoJSON(%s,%s,%s,%s)' % (
            second['version'], self.expand(first, query_env=query_env),
            second['precision'], second['options'])

    def st_astext(self, first, query_env={}):
        return 'ST_AsText(%s)' % self.expand(first, query_env=query_env)

    def st_aswkb(self, first, query_env={}):
        return '%s' % self.expand(first, query_env=query_env)

    def st_x(self, first, query_env={}):
        return 'ST_X(%s)' % (self.expand(first, query_env=query_env))

    def st_y(self, first, query_env={}):
        return 'ST_Y(%s)' % (self.expand(first, query_env=query_env))

    def st_contains(self, first, second, query_env={}):
        return 'ST_Contains(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_distance(self, first, second, query_env={}):
        return 'ST_Distance(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_equals(self, first, second, query_env={}):
        return 'ST_Equals(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_intersects(self, first, second, query_env={}):
        return 'ST_Intersects(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_overlaps(self, first, second, query_env={}):
        return 'ST_Overlaps(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_simplify(self, first, second, query_env={}):
        return 'ST_Simplify(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, 'double', query_env=query_env))

    def st_simplifypreservetopology(self, first, second, query_env={}):
        return 'ST_SimplifyPreserveTopology(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, 'double', query_env=query_env))

    def st_touches(self, first, second, query_env={}):
        return 'ST_Touches(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_within(self, first, second, query_env={}):
        return 'ST_Within(%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def st_dwithin(self, first, tup, query_env={}):
        return 'ST_DWithin(%s,%s,%s)' % (
            self.expand(first, query_env=query_env),
            self.expand(tup[0], first.type, query_env=query_env),
            self.expand(tup[1], 'double', query_env=query_env))

    def st_transform(self, first, second, query_env={}):
    	# The SRID argument can be provided as an integer SRID or a Proj4 string
        if isinstance(second, int):
            return 'ST_Transform(%s,%s)' % (
                self.expand(first, query_env=query_env),
                self.expand(second, 'integer', query_env=query_env))
        else:
            return 'ST_Transform(%s,%s)' % (
                self.expand(first, query_env=query_env),
                self.expand(second, 'string', query_env=query_env))

    @register_expression('doy')
    def extract_doy(self, expr):
        return Expression(expr.db, self.extract, expr, 'doy', 'integer')

    @register_expression('dow')
    def extract_dow(self, expr):
        return Expression(expr.db, self.extract, expr, 'dow', 'integer')

    @register_expression('isodow')
    def extract_isodow(self, expr):
        return Expression(expr.db, self.extract, expr, 'isodow', 'integer')

    @register_expression('isoyear')
    def extract_isoyear(self, expr):
        return Expression(expr.db, self.extract, expr, 'isoyear', 'integer')

    @register_expression('quarter')
    def extract_quarter(self, expr):
        return Expression(expr.db, self.extract, expr, 'quarter', 'integer')

    @register_expression('week')
    def extract_week(self, expr):
        return Expression(expr.db, self.extract, expr, 'week', 'integer')

    @register_expression('decade')
    def extract_decade(self, expr):
        return Expression(expr.db, self.extract, expr, 'decade', 'integer')

    @register_expression('century')
    def extract_century(self, expr):
        return Expression(expr.db, self.extract, expr, 'century', 'integer')

    @register_expression('millenium')
    def extract_millenium(self, expr):
        return Expression(expr.db, self.extract, expr, 'millenium', 'integer')


class PostgreDialectJSON(PostgreDialect):
    @sqltype_for('json')
    def type_json(self):
        return 'JSON'

    @sqltype_for('jsonb')
    def type_jsonb(self):
        return 'JSONB'

    def st_astext(self, first, query_env={}):
        return 'ST_AsText(%s)' % self.expand(first, query_env=query_env)
    
    def st_asgeojson(self, first, second, query_env={}):
        return 'ST_AsGeoJSON(%s,%s,%s,%s)' % (
            second['version'], self.expand(first, query_env=query_env),
            second['precision'], second['options'])

    def json_key(self, first, key, query_env=None):
        """ Get the json in key which you can use for more queries """
        if isinstance(key, basestring):
            key = self.expand(key, 'string', query_env=query_env)
        elif not isinstance(key, integer_types):
            raise TypeError('Key must be a string or int')
        return "%s->%s" % (self.expand(first, query_env=query_env or {}), key)

    def json_key_value(self, first, key, query_env=None):
        """ Get the value int or text in key """
        if isinstance(key, basestring):
            key = self.expand(key, 'string', query_env=query_env)
        elif isinstance(key, integer_types):
            key = self.expand(key, 'integer', query_env=query_env)
        else:
            raise TypeError('Key must be a string or int')
        return "%s->>%s" % (self.expand(first, query_env=query_env or {}), key)

    def json_path(self, first, path, query_env=None):
        """ Get the json in path which you can use for more queries """
        return "%s#>'%s'" % (self.expand(first, query_env=query_env or {}), path)

    def json_path_value(self, first, path, query_env=None):
        """ Get the json in path which you can use for more queries """
        return "%s#>>'%s'" % (self.expand(first, query_env=query_env or {}), path)

    # JSON Queries
    def json_contains(self, first, jsonvalue, query_env=None):
        # requires jsonb, value is json e.g. '{"country": "Peru"}'
        return "%s::jsonb@>'%s'::jsonb" % (self.expand(first, query_env=query_env or {}), jsonvalue)


@dialects.register_for(PostgreNew)
class PostgreDialectArrays(PostgreDialect):
    @sqltype_for('list:integer')
    def type_list_integer(self):
        return 'BIGINT[]'

    @sqltype_for('list:string')
    def type_list_string(self):
        return 'TEXT[]'

    @sqltype_for('list:reference')
    def type_list_reference(self):
        return 'BIGINT[]'

    def any(self, val, query_env={}):
        return "ANY(%s)" % self.expand(val, query_env=query_env)

    def contains(self, first, second, case_sensitive=True, query_env={}):
        if first.type.startswith('list:'):
            f = self.expand(second, 'string', query_env=query_env)
            s = self.any(first, query_env)
            if case_sensitive is True:
                return self.eq(f, s)
            return self.ilike(f, s, escape='\\', query_env=query_env)
        return super(PostgreDialectArrays, self).contains(
            first, second, case_sensitive=case_sensitive, query_env=query_env)

    def ilike(self, first, second, escape=None, query_env={}):
        if first and 'type' not in first:
            args = (first, self.expand(second, query_env=query_env))
            return '(%s ILIKE %s)' % args
        return super(PostgreDialectArrays, self).ilike(
            first, second, escape=escape, query_env=query_env)

    def eq(self, first, second=None, query_env={}):
        if first and 'type' not in first:
            return '(%s = %s)' % (first,
                self.expand(second, query_env=query_env))
        return super(PostgreDialectArrays, self).eq(first, second, query_env)


class PostgreDialectArraysJSON(PostgreDialectArrays):
    @sqltype_for('json')
    def type_json(self):
        return 'JSON'

    @sqltype_for('jsonb')
    def type_jsonb(self):
        return 'JSONB'


@dialects.register_for(PostgreBoolean)
class PostgreDialectBoolean(PostgreDialectArrays):
    @sqltype_for('boolean')
    def type_boolean(self):
        return 'BOOLEAN'


class PostgreDialectBooleanJSON(PostgreDialectBoolean):
    @sqltype_for('json')
    def type_json(self):
        return 'JSON'

    @sqltype_for('jsonb')
    def type_jsonb(self):
        return 'JSONB'



# -*- coding: utf-8 -*-

try:
    from new import classobj
    from google.appengine.ext import db as gae
    from google.appengine.ext import ndb
    from google.appengine.api import namespace_manager, rdbms
    from google.appengine.api.datastore_types import Key  # for belongs on ID
    from google.appengine.ext.ndb.polymodel import PolyModel as NDBPolyModel
except ImportError:
    gae = None
    Key = None



from ..helpers.classes import FakeCursor, SQLALL
from ..helpers.methods import uuid2int
from ..objects import Query, Field
from .base import NoSQLAdapter, SQLAdapter
from . import adapters


@adapters.register_for('couchdb')
class CouchDB(NoSQLAdapter):
    dbengine = 'couchdb'
    drivers = ('couchdb',)

    uploads_in_blob = True

    def _initialize_(self, do_connect):
        super(CouchDB, self)._initialize_(do_connect)
        self.ruri = 'http://' + self.uri[10:]
        self.db_codec = 'UTF-8'

    def connector(self):
        conn = self.driver.Server(self.ruri, **self.driver_args)
        conn.cursor = lambda: FakeCursor()
        conn.close = lambda: None
        conn.commit = lambda: None
        return conn

    def create_table(self, table, migrate=True, fake_migrate=False,
                     polymodel=None):
        if migrate:
            try:
                self.connection.create(table._tablename)
            except:
                pass
        super(CouchDB, self).create_table(
            table, migrate, fake_migrate, polymodel)

    def _expand(self, expression, field_type=None, query_env={}):
        if isinstance(expression, Field):
            if expression.type == 'id':
                return "%s._id" % expression.tablename
        return SQLAdapter._expand(self, expression, field_type,
            query_env=query_env)

    def insert(self, table, fields):
        rid = uuid2int(self.db.uuid())
        ctable = self.connection[table._tablename]
        values = dict((k.name, self.represent(v, k.type)) for k, v in fields)
        values['_id'] = str(rid)
        ctable.save(values)
        return rid

    @staticmethod
    def _make_id_field(field_name):
        return field_name == 'id' and '_id' or field_name

    def _select(self, query, fields, left=False, join=False, distinct=False,
                orderby=False, groupby=False, having=False, limitby=False,
                orderby_on_limitby=True, for_update=False, outer_scoped=[],
                required=None, cache=None, cacheable=None, processor=None):
        if not isinstance(query, Query):
            raise SyntaxError("Not Supported")

        new_fields = []
        for item in fields:
            if isinstance(item, SQLALL):
                new_fields += item._table
            else:
                new_fields.append(item)

        fields = new_fields
        tablename = self.get_table(query)._tablename
        fieldnames = [f.name for f in (fields or self.db[tablename])]
        colnames = [
            '%s.%s' % (tablename, fieldname) for fieldname in fieldnames]
        fields = ','.join(
            ['%s.%s' % (tablename, self._make_id_field(f)) for f in fieldnames]
        )
        fn = '(function(%(t)s){if(%(query)s)emit(%(order)s,[%(fields)s]);})' \
            % dict(
                t=tablename, query=self.expand(query),
                order='%s._id' % tablename, fields=fields)
        return fn, colnames

    def select(self, query, fields, attributes):
        fn, colnames = self._select(query, fields, attributes)
        tablename = colnames[0].split('.')[0]
        ctable = self.connection[tablename]
        rows = [cols['value'] for cols in ctable.query(fn)]
        processor = attributes.get('processor', self.parse)
        return processor(rows, fields, colnames, False)

    def update(self, table, query, fields):
        from ..drivers import couchdb
        if not isinstance(query, Query):
            raise SyntaxError("Not Supported")
        if query.first.type == 'id' and query.op == self.dialect.eq:
            rid = query.second
            tablename = query.first.tablename
            ctable = self.connection[tablename]
            try:
                doc = ctable[str(rid)]
                for key, value in fields:
                    doc[key.name] = self.represent(
                        value, self.db[tablename][key.name].type)
                ctable.save(doc)
                return 1
            except couchdb.http.ResourceNotFound:
                return 0
        tablename = self.get_table(query)._tablename
        rows = self.select(query, [self.db[tablename]._id], {})
        ctable = self.connection[tablename]
        table = self.db[tablename]
        for row in rows:
            doc = ctable[str(row.id)]
            for key, value in fields:
                doc[key.name] = self.represent(value, table[key.name].type)
            ctable.save(doc)
        return len(rows)

    def count(self, query, distinct=None):
        if distinct:
            raise RuntimeError("COUNT DISTINCT not supported")
        if not isinstance(query, Query):
            raise SyntaxError("Not Supported")
        tablename = self.get_table(query)._tablename
        rows = self.select(query, [self.db[tablename]._id], {})
        return len(rows)

    def delete(self, table, query):
        from ..drivers import couchdb
        if not isinstance(query, Query):
            raise SyntaxError("Not Supported")
        if query.first.type == 'id' and query.op == self.eq:
            rid = query.second
            tablename = query.first.tablename
            assert(tablename == query.first.tablename)
            ctable = self.connection[tablename]
            try:
                del ctable[str(rid)]
                return 1
            except couchdb.http.ResourceNotFound:
                return 0
        tablename = self.get_table(query)._tablename
        rows = self.select(query, [self.db[tablename]._id], {})
        ctable = self.connection[tablename]
        for row in rows:
            del ctable[str(row.id)]
        return len(rows)



import re
from .._compat import PY2, iteritems, integer_types, to_unicode, long
from .._globals import IDENTITY
from .base import SQLAdapter
from ..utils import split_uri_args
from . import adapters, with_connection_or_raise

class Slicer(object):
    def rowslice(self, rows, minimum=0, maximum=None):
        if maximum is None:
            return rows[minimum:]
        return rows[minimum:maximum]


class MSSQL(SQLAdapter):
    dbengine = 'mssql'
    drivers = ('pyodbc',)

    REGEX_DSN = '^.+$'
    REGEX_URI = \
         '^(?P<user>[^:@]+)(:(?P<password>[^@]*))?' \
        r'@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?' \
         '/(?P<db>[^?]+)' \
        r'(\?(?P<uriargs>.*))?$'

    def __init__(self, db, uri, pool_size=0, folder=None, db_codec='UTF-8',
                 credential_decoder=IDENTITY, driver_args={},
                 adapter_args={}, do_connect=True, srid=4326,
                 after_connection=None):
        self.srid = srid
        super(MSSQL, self).__init__(
            db, uri, pool_size, folder, db_codec, credential_decoder,
            driver_args, adapter_args, do_connect, after_connection)

    def _initialize_(self, do_connect):
        super(MSSQL, self)._initialize_(do_connect)
        ruri = self.uri.split('://', 1)[1]
        if '@' not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError(
                    "Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group('user'))
            password = self.credential_decoder(m.group('password'))
            if password is None:
                password = ''
            host = m.group('host')
            db = m.group('db')
            port = m.group('port') or '1433'
            # Parse the optional uri name-value arg pairs after the '?'
            # (in the form of arg1=value1&arg2=value2&...)
            # (drivers like FreeTDS insist on uppercase parameter keys)
            argsdict = {'DRIVER': '{SQL Server}'}
            uriargs = m.group('uriargs')
            if uriargs:
                for argkey, argvalue in split_uri_args(
                        uriargs, separators='&', need_equal=True).items():
                    argsdict[argkey.upper()] = argvalue
            uriargs = ';'.join([
                '%s=%s' % (ak, av) for (ak, av) in iteritems(argsdict)])
            self.dsn = 'SERVER=%s;PORT=%s;DATABASE=%s;UID=%s;PWD=%s;%s' \
                % (host, port, db, user, password, uriargs)

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)

    def lastrowid(self, table):
        self.execute('SELECT SCOPE_IDENTITY();')
        return long(self.cursor.fetchone()[0])


@adapters.register_for('mssql')
class MSSQL1(MSSQL, Slicer):
    pass


@adapters.register_for('mssql3')
class MSSQL3(MSSQL):
    pass


@adapters.register_for('mssql4')
class MSSQL4(MSSQL):
    pass


class MSSQLN(MSSQL):
    def represent(self, obj, field_type):
        rv = super(MSSQLN, self).represent(obj, field_type)
        if field_type in ('string', 'text', 'json') and rv.startswith("'"):
            rv = 'N' + rv
        return rv

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        if PY2:
            args = list(args)
            args[0] = to_unicode(args[0])
        return super(MSSQLN, self).execute(*args, **kwargs)


@adapters.register_for('mssqln', 'mssql2')
class MSSQL1N(MSSQLN, Slicer):
    pass


@adapters.register_for('mssql3n')
class MSSQL3N(MSSQLN):
    pass


@adapters.register_for('mssql4n')
class MSSQL4N(MSSQLN):
    pass


@adapters.register_for('vertica')
class Vertica(MSSQL1):
    def lastrowid(self, table):
        self.execute('SELECT SCOPE_IDENTITY();')
        return long(self.cursor.fetchone()[0])


@adapters.register_for('sybase')
class Sybase(MSSQL1):
    dbengine = 'sybase'

    def _initialize_(self, do_connect):
        super(MSSQL, self)._initialize_(do_connect)
        ruri = self.uri.split('://', 1)[1]
        if '@' not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError(
                    "Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group('user'))
            password = self.credential_decoder(m.group('password'))
            if password is None:
                password = ''
            host = m.group('host')
            db = m.group('db')
            port = m.group('port') or '1433'
            self.dsn = 'sybase:host=%s:%s;dbname=%s' % (host, port, db)
            self.driver_args.update(
                user=self.credential_decoder(user),
                passwd=self.credential_decoder(password))



import collections
import copy
import datetime
import fnmatch
import functools
import re

__version__ = '0.1'

__all__ = ['RestAPI', 'Policy', 'ALLOW_ALL_POLICY', 'DENY_ALL_POLICY']

MAX_LIMIT = 1000

class PolicyViolation(ValueError): pass
class InvalidFormat(ValueError): pass
class NotFound(ValueError): pass

def maybe_call(value):
    return value() if callable(value) else value


def error_wrapper(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        data = {}
        try:
            data = func(*args, **kwargs)
            if not data.get('errors'):
                data['status'] = 'success'
                data['code'] = 200
            else:
                data['status'] = 'error'
                data['message'] = 'Validation Errors'
                data['code'] = 422
        except PolicyViolation as e:
            data['status'] = 'error'
            data['message'] = str(e)
            data['code'] = 401
        except NotFound as e:
            data['status'] = 'error'
            data['message'] = str(e)
            data['code'] = 404
        except (InvalidFormat, KeyError, ValueError) as e:
            data['status'] = 'error'
            data['message'] = str(e)
            data['code'] = 400
        finally:
            data['timestamp'] = datetime.datetime.utcnow().isoformat()
            data['api_version'] = __version__
        return data
    return wrapper


class Policy(object):

    model = {
        'POST': {'authorize':False, 'fields':None},
        'PUT': {'authorize':False, 'fields':None},
        'DELETE': {'authorize':False},
        'GET': {'authorize':False, 'fields':None, 'query':None, 'allowed_patterns':[], 'denied_patterns':[], 'limit': MAX_LIMIT, 'allow_lookup': False}
        }

    def __init__(self):
        self.info = {}

    def set(self, tablename, method, **attributes):
        method = method.upper()
        if not method in self.model or any(key not in self.model[method] for key in attributes):
            raise InvalidFormat('Invalid policy format')
        if not tablename in self.info:
            self.info[tablename] = copy.deepcopy(self.model)
        self.info[tablename][method].update(attributes)

    def get(self, tablename, method, name):
        policy = self.info.get(tablename) or self.info.get('*')
        if not policy:
            raise PolicyViolation('No policy for this object')
        return maybe_call(policy[method][name])

    def check_if_allowed(self, method, tablename, id=None, get_vars=None, post_vars=None, exceptions=True):
        get_vars = get_vars or {}
        post_vars = post_vars or {}
        policy = self.info.get(tablename) or self.info.get('*')
        if not policy:
            if exceptions:
                raise PolicyViolation('No policy for this object')
            return False
        policy = policy.get(method.upper())
        if not policy:
            if exceptions:
                raise PolicyViolation('No policy for this method')
            return False
        authorize = policy.get('authorize')
        if authorize is False or (callable(authorize) and not authorize(tablename, id, get_vars, post_vars)):
            if exceptions:
                raise PolicyViolation('Not authorized')
            return False
        for key in get_vars:
            if any(fnmatch.fnmatch(key, p) for p in policy['denied_patterns']):
                if exceptions:
                    raise PolicyViolation('Pattern is not allowed')
                return False
            allowed_patterns = policy['allowed_patterns']
            if '**' not in allowed_patterns and not any(fnmatch.fnmatch(key, p) for p in allowed_patterns):
                if exceptions:
                    raise PolicyViolation('Pattern is not explicitely allowed')
                return False
        return True

    def check_if_lookup_allowed(self, tablename, exceptions=True):
        policy = self.info.get(tablename) or self.info.get('*')
        if not policy:
            if exceptions:
                raise PolicyViolation('No policy for this object')
            return False
        policy = policy.get('GET')
        if not policy:
            if exceptions:
                raise PolicyViolation('No policy for this method')
            return False
        if policy.get('allow_lookup'):
            return True
        return False

    def allowed_fieldnames(self, table, method='GET'):
        method = method.upper()
        policy = self.info.get(table._tablename) or self.info.get('*')
        policy = policy[method]
        allowed_fieldnames = policy['fields']
        if not allowed_fieldnames:
            allowed_fieldnames = [
                f.name for f in table
                if (method == 'GET' and maybe_call(f.readable))
                or (method != 'GET' and maybe_call(f.writable))]
        return allowed_fieldnames

    def check_fieldnames(self, table, fieldnames, method='GET'):
        allowed_fieldnames = self.allowed_fieldnames(table, method)
        invalid_fieldnames = set(fieldnames) - set(allowed_fieldnames)
        if invalid_fieldnames:
            raise InvalidFormat('Invalid fields: %s' % list(invalid_fieldnames))


DENY_ALL_POLICY = Policy()
ALLOW_ALL_POLICY = Policy()
ALLOW_ALL_POLICY.set(tablename='*', method='GET', authorize=True, allowed_patterns=['**'],allow_lookup=True)
ALLOW_ALL_POLICY.set(tablename='*', method='POST', authorize=True)
ALLOW_ALL_POLICY.set(tablename='*', method='PUT', authorize=True)
ALLOW_ALL_POLICY.set(tablename='*', method='DELETE', authorize=True)

class RestAPI(object):

    re_table_and_fields = re.compile(r'\w+([\w+(,\w+)+])?')
    re_lookups = re.compile(r'((\w*\!?\:)?(\w+(\[\w+(,\w+)*\])?)(\.\w+(\[\w+(,\w+)*\])?)*)')
    re_no_brackets = re.compile(r'\[.*?\]')

    def __init__(self, db, policy):
        self.db = db
        self.policy = policy

    @error_wrapper
    def __call__(self, method, tablename, id=None, get_vars=None, post_vars=None):
        method = method.upper()
        get_vars = get_vars or {}
        post_vars = post_vars or {}
        # validate incoming request
        if not tablename in self.db.tables:
            raise InvalidFormat('Invalid table name: %s' % tablename)
        if self.policy:
            self.policy.check_if_allowed(method, tablename, id, get_vars, post_vars)
            if method in ['POST', 'PUT']:
                self.policy.check_fieldnames(self.db[tablename], post_vars.keys(), method)
        # apply rules
        if method == 'GET':
            if id:
                get_vars['id.eq'] = id
            return self.search(tablename, get_vars)
        elif method == 'POST':
            table =  self.db[tablename]
            return table.validate_and_insert(**post_vars).as_dict()
        elif method == 'PUT':
            id = id or post_vars['id']
            if not id:
                raise InvalidFormat('No item id specified')
            table =  self.db[tablename]
            data = self.db(table.id == id).validate_and_update(**post_vars).as_dict()
            if not data.get('updated'):
                raise NotFound('Item not found')
            return data
        elif method == 'DELETE':
            id = id or post_vars['id']
            if not id:
                raise InvalidFormat('No item id specified')
            table = self.db[tablename]
            deleted = self.db(table.id == id).delete()
            if not deleted:
                raise NotFound('Item not found')
            return {'deleted': deleted}

    def table_model(self, table, fieldnames):
        """ converts a table into its form template """
        items = []
        fields = post_fields = put_fields = table.fields
        if self.policy:
            fields = self.policy.allowed_fieldnames(table, method='GET')
            put_fields = self.policy.allowed_fieldnames(table, method='PUT')
            post_fields = self.policy.allowed_fieldnames(table, method='POST')
        for fieldname in fields:
            if fieldnames and not fieldname in fieldnames:
                continue
            field = table[fieldname]
            if not field.readable and not field.writable:
                continue
            item = {'name': field.name, 'label': field.label}
            # https://github.com/collection-json/extensions/blob/master/template-validation.md
            item['default'] = field.default() if callable(field.default) else field.default
            parts = field.type.split()
            item['type'] = parts[0]
            if len(parts)>1:
                item['references'] = parts[1]
            if hasattr(field,'regex'):
                item['regex'] = field.regex
            item['required'] = field.required
            item['unique'] = field.unique
            item['post_writable'] = field.name in post_fields
            item['put_writable'] = field.name in put_fields
            item['options'] = field.options
            if field.type == 'id':
                item['referenced_by'] = ['%s.%s' % (f._tablename, f.name)
                                         for f in table._referenced_by
                                         if self.policy and
                                         self.policy.check_if_allowed('GET', f._tablename, 
                                                                      exceptions=False)]
            items.append(item)
        return items

    @staticmethod
    def make_query(field, condition, value):
        expression = {
            'eq': lambda: field == value,
            'ne': lambda: field == value,
            'lt': lambda: field < value,
            'gt': lambda: field > value,
            'le': lambda: field <= value,
            'ge': lambda: field >= value,
            'startswith': lambda: field.startswith(str(value)),
            'in': lambda: field.belongs(value.split(',') if isinstance(value,str) else list(value)),
            'contains': lambda: field.contains(value),
            }
        return expression[condition]()

    @staticmethod
    def parse_table_and_fields(text):
        if not RestAPI.re_table_and_fields.match(text):
            raise ValueError
        parts = text.split('[')
        if len(parts) == 1:
            return parts[0], []
        elif len(parts) == 2:
            return parts[0], parts[1][:-1].split(',')

    def search(self, tname, vars):

        def check_table_permission(tablename):
            if self.policy:
                self.policy.check_if_allowed('GET', tablename)

        def check_table_lookup_permission(tablename):
            if self.policy:
                self.policy.check_if_lookup_allowed(tablename)

        def filter_fieldnames(table, fieldnames):
            if self.policy:
                if fieldnames:
                    self.policy.check_fieldnames(table, fieldnames)
                else:
                    fieldnames = self.policy.allowed_fieldnames(table)
            elif not fieldnames:
                fieldnames = table.fields
            return fieldnames

        db = self.db
        tname, tfieldnames = RestAPI.parse_table_and_fields(tname)
        check_table_permission(tname)
        tfieldnames = filter_fieldnames(db[tname], tfieldnames)
        query = []
        offset = 0
        limit = 100
        model = False
        table = db[tname]
        queries = []
        if self.policy:
            common_query = self.policy.get(tname, 'GET', 'query')
            if common_query:
                queries.append(common_query)
        hop1 = collections.defaultdict(list)
        hop2 = collections.defaultdict(list)
        hop3 = collections.defaultdict(list)
        model_fieldnames = tfieldnames
        lookup = {}
        orderby = None
        for key, value in vars.items():
            if key == '@offset':
                offset = int(value)
            elif key == '@limit':
                limit = min(int(value), self.policy.get(tname, 'GET', 'limit') if self.policy else MAX_LIMIT)
            elif key == '@order':
                orderby = [~table[f[1:]] if f[:1] == '~' else table[f] for f in value.split(',')
                          if (f[1:] if f[:1] == '~' else f) in table.fields] or None
            elif key == '@lookup':
                lookup = {item[0]: {} for item in RestAPI.re_lookups.findall(value)}
            elif key == '@model':
                model = str(value).lower()[:1] == 't'
            else:
                key_parts = key.rsplit('.')
                if not key_parts[-1] in ('eq','ne','gt','lt','ge','le','startswith'):
                    key_parts.append('eq')
                is_negated = key_parts[0] == 'not'
                if is_negated:
                    key_parts = key_parts[1:]
                key, condition = key_parts[:-1], key_parts[-1]
                if len(key) == 1: # example: name.eq=='Chair'
                    query = self.make_query(table[key[0]], condition, value)
                    queries.append(query if not is_negated else ~query)
                elif len(key) == 2: # example: color.name.eq=='red'
                    hop1[is_negated, key[0]].append((key[1], condition, value))
                elif len(key) == 3: # example: a.rel.desc.eq=='above'
                    hop2[is_negated, key[0], key[1]].append((key[2], condition, value))
                elif len(key) == 4: # example: a.rel.b.name.eq == 'Table'
                    hop3[is_negated, key[0], key[1], key[2]].append((key[3], condition, value))

        for item in hop1:
            is_negated, fieldname = item
            ref_tablename = table[fieldname].type.split(' ')[1]
            ref_table = db[ref_tablename]
            subqueries = [self.make_query(ref_table[k], c, v) for k,c,v in hop1[item]]
            subquery = functools.reduce(lambda a,b: a&b, subqueries)
            query = table[fieldname].belongs(db(subquery)._select(ref_table.id))
            queries.append(query if not is_negated else ~query)

        for item in hop2:
            is_negated, linkfield, linktable = item
            ref_table = db[linktable]
            subqueries = [self.make_query(ref_table[k], c, v) for k, c, v in hop2[item]]
            subquery = functools.reduce(lambda a,b: a&b, subqueries)
            query = table.id.belongs(db(subquery)._select(ref_table[linkfield]))
            queries.append(query if not is_negated else ~query)

        for item in hop3:
            is_negated, linkfield, linktable, otherfield = item
            ref_table = db[linktable]
            ref_ref_tablename = ref_table[otherfield].type.split(' ')[1]
            ref_ref_table = db[ref_ref_tablename]
            subqueries = [self.make_query(ref_ref_table[k], c, v) for k,c,v in hop3[item]]
            subquery = functools.reduce(lambda a, b: a&b, subqueries)
            subquery &= ref_ref_table.id == ref_table[otherfield]
            query = table.id.belongs(db(subquery)._select(ref_table[linkfield], groupby=ref_table[linkfield]))
            queries.append(query if not is_negated else ~query)

        if not queries:
            queries.append(table)

        query = functools.reduce(lambda a, b: a&b, queries)
        tfields = [table[tfieldname] for tfieldname in tfieldnames]
        rows = db(query).select(*tfields, limitby=(offset, limit + offset), orderby=orderby)

        lookup_map = {}
        for key in list(lookup.keys()):
            name, key = key.split(':') if ':' in key else ('', key)
            clean_key = RestAPI.re_no_brackets.sub('', key)
            lookup_map[clean_key] = {'name': name.rstrip('!') or clean_key,
                                     'collapsed': name.endswith('!')}
            key = key.split('.')

            if len(key) == 1:
                key, tfieldnames = RestAPI.parse_table_and_fields(key[0])
                ref_tablename = table[key].type.split(' ')[1]
                ref_table = db[ref_tablename]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                check_table_lookup_permission(ref_tablename)
                ids = [row[key] for row in rows]
                tfields = [ref_table[tfieldname] for tfieldname in tfieldnames]
                if not 'id' in tfieldnames:
                    tfields.append(ref_table['id'])
                drows = db(ref_table.id.belongs(ids)).select(*tfields).as_dict()
                if tfieldnames and not 'id' in tfieldnames:
                    for row in drows.values():
                        del row['id']
                lkey, collapsed = lookup_map[key]['name'], lookup_map[key]['collapsed']
                for row in rows:
                    new_row = drows.get(row[key])
                    if new_row and collapsed:
                        del row[key]
                        for rkey in new_row:
                            row[lkey + '_' + rkey] = new_row[rkey]
                    else:
                        row[lkey] = new_row

            elif len(key) == 2:
                lfield, key = key
                key, tfieldnames = RestAPI.parse_table_and_fields(key)
                check_table_lookup_permission(key)
                ref_table = db[key]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                ids = [row['id'] for row in rows]
                tfields = [ref_table[tfieldname] for tfieldname in tfieldnames]
                if not lfield in tfieldnames:
                    tfields.append(ref_table[lfield])
                lrows = db(ref_table[lfield].belongs(ids)).select(*tfields)
                drows = collections.defaultdict(list)
                for row in lrows:
                    row = row.as_dict()
                    drows[row[lfield]].append(row)
                    if not lfield in tfieldnames:
                        del row[lfield]
                lkey = lookup_map[lfield + '.' + key]['name']
                for row in rows:
                    row[lkey] = drows.get(row.id, [])

            elif len(key) == 3:
                lfield, key, rfield = key
                key, tfieldnames = RestAPI.parse_table_and_fields(key)
                rfield, tfieldnames2 = RestAPI.parse_table_and_fields(rfield)
                check_table_lookup_permission(key)
                ref_table = db[key]
                ref_ref_tablename = ref_table[rfield].type.split(' ')[1]
                check_table_lookup_permission(ref_ref_tablename)
                ref_ref_table = db[ref_ref_tablename]
                tfieldnames = filter_fieldnames(ref_table, tfieldnames)
                tfieldnames2 = filter_fieldnames(ref_ref_table, tfieldnames2)
                ids = [row['id'] for row in rows]
                tfields = [ref_table[tfieldname] for tfieldname in tfieldnames]
                if not lfield in tfieldnames:
                    tfields.append(ref_table[lfield])
                if not rfield in tfieldnames:
                    tfields.append(ref_table[rfield])
                tfields += [ref_ref_table[tfieldname] for tfieldname in tfieldnames2]
                left = ref_ref_table.on(ref_table[rfield]==ref_ref_table['id'])
                lrows = db(ref_table[lfield].belongs(ids)).select(*tfields, left=left)
                drows = collections.defaultdict(list)
                lkey = lfield + '.' + key + '.' + rfield
                lkey, collapsed = lookup_map[lkey]['name'], lookup_map[lkey]['collapsed']
                for row in lrows:
                    row = row.as_dict()
                    new_row = row[key]
                    lfield_value, rfield_value = new_row[lfield], new_row[rfield]
                    if not lfield in tfieldnames:
                        del new_row[lfield]
                    if not rfield in tfieldnames:
                        del new_row[rfield]
                    if collapsed:
                        new_row.update(row[ref_ref_tablename])
                    else:
                        new_row[rfield] = row[ref_ref_tablename]
                    drows[lfield_value].append(new_row)
                for row in rows:
                    row[lkey] = drows.get(row.id, [])

        response = {}
        response['items'] = rows.as_list()
        if offset == 0:
            response['count'] = db(query).count()
        if model:
            response['model'] = self.table_model(table, model_fieldnames)
        return response



class Dispatcher(object):
    namespace = "dispatcher"

    def __init__(self, namespace=None):
        self._registry_ = {}
        if namespace:
            self.namespace = namespace

    def register_for(self, target):
        def wrap(dispatch_class):
            self._registry_[target] = dispatch_class
            return dispatch_class
        return wrap

    def get_for(self, obj):
        targets = type(obj).__mro__
        for target in targets:
            if target in self._registry_:
                return self._registry_[target](obj)
        else:
            raise ValueError(
                "no %s found for object: %s" % (self.namespace, obj))


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self.cursor = connection.cursor()
        self.release()

    def lock(self):
        self.available = False

    def release(self):
        self.available = True



import re
from .._compat import PY2, basestring
from ..adapters.mongo import Mongo
from ..exceptions import NotOnNOSQLError
from ..objects import Field
from .base import NoSQLDialect
from . import dialects

_aggregate_map = {
    'SUM': '$sum',
    'MAX': '$max',
    'MIN': '$min',
    'AVG': '$avg',
}

_extract_map = {
    'dayofyear':    '$dayOfYear',
    'day':          '$dayOfMonth',
    'dayofweek':    '$dayOfWeek',
    'year':         '$year',
    'month':        '$month',
    'week':         '$week',
    'hour':         '$hour',
    'minute':       '$minute',
    'second':       '$second',
    'millisecond':  '$millisecond',
    'string':       '$dateToString',
}


def needs_aggregation_pipeline(f):
    def wrap(self, first, *args, **kwargs):
        self.adapter._parse_data(first, 'pipeline', True)
        if len(args) > 0:
            self.adapter._parse_data(args[0], 'pipeline', True)
        return f(self, first, *args, **kwargs)
    return wrap


def validate_second(f):
    def wrap(*args, **kwargs):
        if len(args) < 3 or args[2] is None:
            raise RuntimeError("Cannot compare %s with None" % args[1])
        return f(*args, **kwargs)
    return wrap


def check_fields_for_cmp(f):
    def wrap(self, first, second=None, *args, **kwargs):
        if (self.adapter._parse_data((first, second), 'pipeline')):
            pipeline = True
        elif not isinstance(first, Field) or self._has_field(second):
            pipeline = True
            self.adapter._parse_data((first, second), 'pipeline', True)
        else:
            pipeline = False
        return f(self, first, second, *args, pipeline=pipeline, **kwargs)
    return wrap


@dialects.register_for(Mongo)
class MongoDialect(NoSQLDialect):
    GROUP_MARK = "__#GROUP#__"
    AS_MARK = "__#AS#__"
    REGEXP_MARK1 = "__#REGEXP_1#__"
    REGEXP_MARK2 = "__#REGEXP_2#__"
    REGEX_SELECT_AS_PARSER = r"'%s': '(\S+)'" % AS_MARK

    @staticmethod
    def _has_field(expression):
        try:
            return expression.has_field
        except AttributeError:
            return False

    def invert(self, first, query_env={}):
        return '-%s' % self.expand(first, query_env=query_env)

    def _not(self, val, query_env={}):
        op = self.expand(val, query_env=query_env)
        op_k = list(op)[0]
        op_body = op[op_k]
        rv = None
        if type(op_body) is list:
            # apply De Morgan law for and/or
            # not(A and B) -> not(A) or not(B)
            # not(A or B)  -> not(A) and not(B)
            not_op = '$and' if op_k == '$or' else '$or'
            rv = {not_op: [self._not(val.first, query_env),
                self._not(val.second, query_env)]}
        else:
            try:
                sub_ops = list(op_body.keys())
                if len(sub_ops) == 1 and sub_ops[0] == '$ne':
                    rv = {op_k: op_body['$ne']}
            except AttributeError:
                rv = {op_k: {'$ne': op_body}}
            if rv is None:
                rv = {op_k: {'$not': op_body}}
        return rv

    def _and(self, first, second, query_env={}):
        # pymongo expects: .find({'$and': [{'x':'1'}, {'y':'2'}]})
        if isinstance(second, bool):
            if second:
                return self.expand(first, query_env=query_env)
            return self.ne(first, first)
        return {'$and': [self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env)]}

    def _or(self, first, second, query_env={}):
        # pymongo expects: .find({'$or': [{'name':'1'}, {'name':'2'}]})
        if isinstance(second, bool):
            if not second:
                return self.expand(first, query_env=query_env)
            return True
        return {'$or': [self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env)]}

    def belongs(self, first, second, query_env={}):
        if isinstance(second, str):
            # this is broken, the only way second is a string is if it has
            # been converted to SQL. This no worky. This might be made to
            # work if adapter._select did not return SQL.
            raise RuntimeError("nested queries not supported")
        items = [self.expand(item, first.type, query_env=query_env)
            for item in second]
        return {self.expand(first, query_env=query_env): {"$in": items}}

    def _cmp_ops_aggregation_pipeline(self, op, first, second, query_env={}):
        try:
            type = first.type
        except:
            type = None
        return {op: [self.expand(first, query_env=query_env),
            self.expand(second, type, query_env=query_env)]}

    @check_fields_for_cmp
    def eq(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$eq', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            self.expand(second, first.type, query_env=query_env)}

    @check_fields_for_cmp
    def ne(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$ne', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            {'$ne': self.expand(second, first.type, query_env=query_env)}}

    @validate_second
    @check_fields_for_cmp
    def lt(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$lt', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            {'$lt': self.expand(second, first.type, query_env=query_env)}}

    @validate_second
    @check_fields_for_cmp
    def lte(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$lte', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            {'$lte': self.expand(second, first.type, query_env=query_env)}}

    @validate_second
    @check_fields_for_cmp
    def gt(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$gt', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            {'$gt': self.expand(second, first.type, query_env=query_env)}}

    @validate_second
    @check_fields_for_cmp
    def gte(self, first, second=None, pipeline=False, query_env={}):
        if pipeline:
            return self._cmp_ops_aggregation_pipeline('$gte', first, second,
                query_env)
        return {self.expand(first, query_env=query_env):
            {'$gte': self.expand(second, first.type, query_env=query_env)}}

    @needs_aggregation_pipeline
    def add(self, first, second, query_env={}):
        op_code = '$add'
        for field in [first, second]:
            try:
                if field.type in ['string', 'text', 'password']:
                    op_code = '$concat'
                    break
            except:
                pass
        return {op_code: [self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env)]}

    @needs_aggregation_pipeline
    def sub(self, first, second, query_env={}):
        return {'$subtract': [
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env)]}

    @needs_aggregation_pipeline
    def mul(self, first, second, query_env={}):
        return {'$multiply': [
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env)]}

    @needs_aggregation_pipeline
    def div(self, first, second, query_env={}):
        return {'$divide': [
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env)]}

    @needs_aggregation_pipeline
    def mod(self, first, second, query_env={}):
        return {'$mod': [
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env)]}

    @needs_aggregation_pipeline
    def aggregate(self, first, what, query_env={}):
        if what == 'ABS':
            return {
                "$cond": [
                    {"$lt": [self.expand(first, query_env=query_env), 0]},
                    {"$subtract": [0, self.expand(first, query_env=query_env)]},
                    self.expand(first, query_env=query_env)]}
        try:
            expanded = {_aggregate_map[what]:
                self.expand(first, query_env=query_env)}
        except KeyError:
            raise NotImplementedError("'%s' not implemented" % what)

        self.adapter._parse_data(first, 'need_group', True)
        return {self.GROUP_MARK: expanded}

    @needs_aggregation_pipeline
    def count(self, first, distinct=None, query_env={}):
        self.adapter._parse_data(first, 'need_group', True)
        if distinct:
            ret = {self.GROUP_MARK: {"$addToSet":
                self.expand(first, query_env=query_env)}}
            if self.adapter.server_version_major >= 2.6:
                # '$size' not present in server versions < 2.6
                ret = {'$size': ret}
            return ret
        return {self.GROUP_MARK: {"$sum": 1}}

    @needs_aggregation_pipeline
    def extract(self, first, what, query_env={}):
        try:
            return {_extract_map[what]: self.expand(first, query_env=query_env)}
        except KeyError:
            raise NotImplementedError("EXTRACT(%s) not implemented" % what)

    @needs_aggregation_pipeline
    def epoch(self, first, query_env={}):
        return {"$divide": [
            {"$subtract": [self.expand(first, query_env=query_env),
                self.adapter.epoch]}, 1000]}

    @needs_aggregation_pipeline
    def case(self, query, true_false, query_env={}):
        return {"$cond": [
            self.expand(query, query_env=query_env),
            self.expand(true_false[0], query_env=query_env),
            self.expand(true_false[1], query_env=query_env)]}

    @needs_aggregation_pipeline
    def _as(self, first, second, query_env={}):
        # put the AS_MARK into the structure.  The 'AS' name will be parsed
        # later from the string of the field name.
        if isinstance(first, Field):
            return [{self.AS_MARK: second},
                self.expand(first, query_env=query_env)]
        else:
            result = self.expand(first, query_env=query_env)
            result[self.AS_MARK] = second
        return result

    # We could implement an option that simulates a full featured SQL
    # database. But I think the option should be set explicit or
    # implemented as another library.
    def on(self, first, second, query_env={}):
        raise NotOnNOSQLError()

    def comma(self, first, second, query_env={}):
        # returns field name lists, to be separated via split(',')
        return '%s,%s' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    # TODO verify full compatibilty with official SQL Like operator
    def _build_like_regex(self, first, second, case_sensitive=True,
                          escape=None, ends_with=False, starts_with=False,
                          whole_string=True, like_wildcards=False,
                          query_env={}):
        base = self.expand(second, 'string', query_env=query_env)
        need_regex = (whole_string or not case_sensitive or starts_with or
                      ends_with or like_wildcards and
                      ('_' in base or '%' in base))
        if not need_regex:
            return base
        expr = re.escape(base)
        if like_wildcards:
            if escape:
                # protect % and _ which are escaped
                expr = expr.replace(escape+'\\%', '%')
                if PY2:
                    expr = expr.replace(escape+'\\_', '_')
                elif escape+'_' in expr:
                    set_aside = str(self.adapter.object_id('<random>'))
                    while set_aside in expr:
                        set_aside = str(self.adapter.object_id('<random>'))
                    expr = expr.replace(escape+'_', set_aside)
                else:
                    set_aside = None
            expr = expr.replace('\\%', '.*')
            if PY2:
                expr = expr.replace('\\_', '.')
            else:
                expr = expr.replace('_', '.')
            if escape:
                # convert to protected % and _
                expr = expr.replace('%', '\\%')
                if PY2:
                    expr = expr.replace('_', '\\_')
                elif set_aside:
                    expr = expr.replace(set_aside, '_')
        if starts_with:
            pattern = '^%s'
        elif ends_with:
            pattern = '%s$'
        elif whole_string:
            pattern = '^%s$'
        else:
            pattern = '%s'
        return self.regexp(first, pattern % expr, case_sensitive, query_env)

    def like(self, first, second, case_sensitive=True, escape=None,
        query_env={}):
        return self._build_like_regex(
            first, second, case_sensitive=case_sensitive, escape=escape,
            like_wildcards=True, query_env=query_env)

    def ilike(self, first, second, escape=None, query_env={}):
        return self.like(first, second, case_sensitive=False, escape=escape,
            query_env=query_env)

    def startswith(self, first, second, query_env={}):
        return self._build_like_regex(first, second, starts_with=True,
            query_env=query_env)

    def endswith(self, first, second, query_env={}):
        return self._build_like_regex(first, second, ends_with=True,
            query_env=query_env)

    # TODO verify full compatibilty with official oracle contains operator
    def contains(self, first, second, case_sensitive=True, query_env={}):
        if isinstance(second, self.adapter.ObjectId):
            ret = {self.expand(first, query_env=query_env): second}
        elif isinstance(second, Field):
            if second.type in ['string', 'text']:
                if isinstance(first, Field):
                    if first.type in ['list:string', 'string', 'text']:
                        ret = {
                            '$where': "this.%s.indexOf(this.%s) > -1" % (
                                first.name, second.name)}
                    else:
                        raise NotImplementedError(
                            "field.CONTAINS() not implemented for field " +
                            "type of '%s'" % first.type)
                else:
                    raise NotImplementedError(
                        "x.CONTAINS() not implemented for x type of '%s'" %
                        type(first))
            elif second.type in ['integer', 'bigint']:
                ret = {
                    '$where': "this.%s.indexOf(this.%s + '') > -1" % (
                        first.name, second.name)}
            else:
                raise NotImplementedError(
                    "CONTAINS(field) not implemented for field type '%s'" %
                    second.type)
        elif isinstance(second, (basestring, int)):
            whole_string = isinstance(first, Field) and \
                first.type == 'list:string'
            ret = self._build_like_regex(
                first, second, case_sensitive=case_sensitive,
                whole_string=whole_string, query_env=query_env)
            # first.type in ('string', 'text', 'json', 'upload')
            # or first.type.startswith('list:'):
        else:
            raise NotImplementedError(
                "CONTAINS() not implemented for type '%s'" % type(second))
        return ret

    @needs_aggregation_pipeline
    def substring(self, field, parameters, query_env={}):
        def parse_parameters(pos0, length):
            """
            The expression object can return these as string based expressions.
            We can't use that so we have to tease it apart.

            These are the possibilities:

              pos0 = '(%s - %d)' % (self.len(), abs(start) - 1)
              pos0 = start + 1

              length = self.len()
              length = '(%s - %d - %s)' % (self.len(), abs(stop) - 1, pos0)
              length = '(%s - %s)' % (stop + 1, pos0)

            Two of these five require the length of the string which is not
            supported by Mongo, so for now these cause an Exception and
            won't reach here.

            If this were to ever be supported it may require a change to
            Expression.__getitem__ so that it either returned the base
            expression to be expanded here, or converted length to a string
            to be parsed back to a call to STRLEN()
            """
            if isinstance(length, basestring):
                return (pos0 - 1, eval(length))
            # take the rest of the string
            return (pos0 - 1, -1)

        parameters = parse_parameters(*parameters)
        return {'$substr': [self.expand(field, query_env=query_env),
            parameters[0], parameters[1]]}

    @needs_aggregation_pipeline
    def lower(self, first, query_env={}):
        return {'$toLower': self.expand(first, query_env=query_env)}

    @needs_aggregation_pipeline
    def upper(self, first, query_env={}):
        return {'$toUpper': self.expand(first, query_env=query_env)}

    def regexp(self, first, second, case_sensitive=True, query_env={}):
        """ MongoDB provides regular expression capabilities for pattern
            matching strings in queries. MongoDB uses Perl compatible
            regular expressions (i.e. 'PCRE') version 8.36 with UTF-8 support.
        """
        if (isinstance(first, Field) and
                first.type in ['integer', 'bigint', 'float', 'double']):
            return {
                '$where': "RegExp('%s').test(this.%s + '')" % (
                    self.expand(second, 'string', query_env=query_env),
                    first.name)}
        expanded_first = self.expand(first, query_env=query_env)
        regex_second = {'$regex': self.expand(second, 'string',
            query_env=query_env)}
        if not case_sensitive:
            regex_second['$options'] = 'i'
        if (self.adapter._parse_data((first, second), 'pipeline')):
            name = str(expanded_first)
            return {
                self.REGEXP_MARK1: {name: expanded_first},
                self.REGEXP_MARK2: {name: regex_second}}
        try:
            return {expanded_first: regex_second}
        except TypeError:
            # if first is not hashable, then will need the pipeline
            self.adapter._parse_data((first, second), 'pipeline', True)
            return {}

    def length(self, first, query_env={}):
        """
        Mongo has committed $strLenBytes, $strLenCP, and $substrCP to $project
        aggregation stage in dev branch V3.3.4

        https://jira.mongodb.org/browse/SERVER-14670
        https://jira.mongodb.org/browse/SERVER-22580
        db.coll.aggregate([{
            $project: {
                byteLength: {$strLenBytes: "$string"},
                cpLength: {$strLenCP: "$string"}
                byteSubstr: {$substrBytes: ["$string", 0, 4]},
                cpSubstr: {$substrCP: ["$string", 0, 4]}
            }
        }])

        https://jira.mongodb.org/browse/SERVER-5319
        https://github.com/afchin/mongo/commit/f52105977e4d0ccb53bdddfb9c4528a3f3c40bdf
        """
        if self.adapter.server_version_major <= 3.2:
            # $strLenBytes not supported by mongo before version 3.4
            raise NotImplementedError()

        # implement here  :-)
        raise NotImplementedError()

    @needs_aggregation_pipeline
    def coalesce(self, first, second, query_env={}):
        if len(second) > 1:
            second = [self.coalesce(second[0], second[1:])]
        return {"$ifNull": [self.expand(first, query_env=query_env),
            self.expand(second[0], query_env=query_env)]}

    @property
    def random(self):
        """ ORDER BY RANDOM()

        Mongo has released the '$sample' pipeline stage in V3.2
        https://docs.mongodb.org/manual/reference/operator/aggregation/sample/

        https://github.com/mongodb/cookbook/blob/master/content/patterns/random-attribute.txt
        http://stackoverflow.com/questions/19412/how-to-request-a-random-row-in-sql
        https://jira.mongodb.org/browse/SERVER-533
        """

        if self.adapter.server_version_major <= 3.0:
            # '$sample' not present until server version 3.2
            raise NotImplementedError()

        # implement here  :-)
        raise NotImplementedError()



import locale
import platform
import re
import sys
from datetime import datetime
from time import mktime
from .._compat import PY2, pjoin
from .base import SQLAdapter
from . import adapters


@adapters.register_for('sqlite', 'sqlite:memory')
class SQLite(SQLAdapter):
    dbengine = 'sqlite'
    drivers = ('sqlite2', 'sqlite3')

    def _initialize_(self, do_connect):
        self.pool_size = 0
        super(SQLite, self)._initialize_(do_connect)
        path_encoding = sys.getfilesystemencoding() \
            or locale.getdefaultlocale()[1] or 'utf8'
        if ':memory' in self.uri.split('://', 1)[0]:
            self.dbpath = ':memory:'
        else:
            self.dbpath = self.uri.split('://', 1)[1]
            if self.dbpath[0] != '/':
                if PY2:
                    self.dbpath = pjoin(
                        self.folder.decode(path_encoding).encode('utf8'),
                        self.dbpath)
                else:
                    self.dbpath = pjoin(self.folder, self.dbpath)
        if 'check_same_thread' not in self.driver_args:
            self.driver_args['check_same_thread'] = False
        if 'detect_types' not in self.driver_args and do_connect:
            self.driver_args['detect_types'] = self.driver.PARSE_DECLTYPES

    def _driver_from_uri(self):
        return None

    def connector(self):
        return self.driver.Connection(self.dbpath, **self.driver_args)

    @staticmethod
    def web2py_extract(lookup, s):
        table = {
            'year': (0, 4), 'month': (5, 7), 'day': (8, 10), 'hour': (11, 13),
            'minute': (14, 16), 'second': (17, 19)}
        try:
            if lookup != 'epoch':
                (i, j) = table[lookup]
                return int(s[i:j])
            else:
                return mktime(
                    datetime.strptime(s, '%Y-%m-%d %H:%M:%S').timetuple())
        except:
            return None

    @staticmethod
    def web2py_regexp(expression, item):
        if item is None:
            return False
        return re.compile(expression).search(item) is not None

    def _register_extract(self):
        self.connection.create_function(
            'web2py_extract', 2, self.web2py_extract)

    def _register_regexp(self):
        self.connection.create_function(
            "REGEXP", 2, self.web2py_regexp)

    def after_connection(self):
        self._register_extract()
        self._register_regexp()
        if self.adapter_args.get('foreign_keys', True):
            self.execute('PRAGMA foreign_keys=ON;')

    def select(self, query, fields, attributes):
        if attributes.get('for_update', False) and 'cache' not in attributes:
            self.execute('BEGIN IMMEDIATE TRANSACTION;')
        return super(SQLite, self).select(query, fields, attributes)

    def delete(self, table, query):
        db = self.db
        deleted = [x[table._id.name] for x in db(query).select(table._id)]
        counter = super(SQLite, self).delete(table, query)
        if counter:
            for field in table._referenced_by:
                if field.type == 'reference ' + table._dalname \
                   and field.ondelete == 'CASCADE':
                    db(field.belongs(deleted)).delete()
        return counter


@adapters.register_for('spatialite', 'spatialite:memory')
class Spatialite(SQLite):
    dbengine = 'spatialite'

    SPATIALLIBS = {
        'Windows': 'mod_spatialite.dll',
        'Linux': 'libspatialite.so',
        'Darwin': 'libspatialite.dylib'
    }

    def after_connections(self):
        self.connection.enable_load_extension(True)
        libspatialite = self.SPATIALLIBS[platform.system()]
        self.execute(r'SELECT load_extension("%s");' % libspatialite)
        super(Spatialite, self).after_connection()


@adapters.register_for('jdbc:sqlite', 'jdbc:sqlite:memory')
class JDBCSQLite(SQLite):
    drivers = ('zxJDBC_sqlite',)

    def connector(self):
        return self.driver.connect(
            self.driver.getConnection('jdbc:sqlite:'+self.dbpath),
            **self.driver_args)

    def after_connection(self):
        self._register_extract()



import datetime
from ..adapters.informix import Informix
from .base import SQLRepresenter
from . import representers


@representers.register_for(Informix)
class InformixRepresenter(SQLRepresenter):
    def exceptions(self, obj, field_type):
        if field_type == 'date':
            if isinstance(obj, (datetime.date, datetime.datetime)):
                obj = obj.isoformat()[:10]
            else:
                obj = str(obj)
            return "to_date('%s','%%Y-%%m-%%d')" % obj
        elif field_type == 'datetime':
            if isinstance(obj, datetime.datetime):
                obj = obj.isoformat()[:19].replace('T', ' ')
            elif isinstance(obj, datetime.date):
                obj = obj.isoformat()[:10]+' 00:00:00'
            else:
                obj = str(obj)
            return "to_date('%s','%%Y-%%m-%%d %%H:%%M:%%S')" % obj
        return None



# -*- coding: utf-8 -*-

from ._compat import unittest
from ._adapt import DEFAULT_URI, drop, IS_MSSQL, IS_IMAP, IS_GAE, IS_TERADATA
from pydal import DAL, Field
from pydal._compat import PY2


@unittest.skipIf(IS_IMAP, "Reference not Null unsupported on IMAP")
class TestReferenceNOTNULL(unittest.TestCase):
    #1:N not null

    def testRun(self):
        for ref, bigint in [('reference', False), ('big-reference', True)]:
            db = DAL(DEFAULT_URI, check_reserved=['all'], bigint_id=bigint)
            if bigint and 'big-id' not in db._adapter.types:
                continue
            db.define_table('tt', Field('vv'))
            db.define_table('ttt', Field('vv'), Field('tt_id', '%s tt' % ref,
                                                      notnull=True))
            self.assertRaises(Exception, db.ttt.insert, vv='pydal')
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Reference Unique unsupported on IMAP")
@unittest.skipIf(IS_GAE, "Reference Unique unsupported on GAE")
class TestReferenceUNIQUE(unittest.TestCase):
    # 1:1 relation

    def testRun(self):
        for ref, bigint in [('reference', False), ('big-reference', True)]:
            db = DAL(DEFAULT_URI, check_reserved=['all'], bigint_id=bigint)
            if bigint and 'big-id' not in db._adapter.types:
                continue
            db.define_table('tt', Field('vv'))
            db.define_table('ttt', Field('vv'),
                            Field('tt_id', '%s tt' % ref, unique=True),
                            Field('tt_uq', 'integer', unique=True))
            id_1 = db.tt.insert(vv='pydal')
            id_2 = db.tt.insert(vv='pydal')
            # Null tt_id
            db.ttt.insert(vv='pydal', tt_uq=1)
            # first insert is OK
            db.ttt.insert(tt_id=id_1, tt_uq=2)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_1, tt_uq=3)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_2, tt_uq=2)
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Reference Unique not Null unsupported on IMAP")
@unittest.skipIf(IS_GAE, "Reference Unique not Null unsupported on GAE")
class TestReferenceUNIQUENotNull(unittest.TestCase):
    # 1:1 relation not null

    def testRun(self):
        for ref, bigint in [('reference', False), ('big-reference', True)]:
            db = DAL(DEFAULT_URI, check_reserved=['all'], bigint_id=bigint)
            if bigint and 'big-id' not in db._adapter.types:
                continue
            db.define_table('tt', Field('vv'))
            db.define_table('ttt', Field('vv'), Field('tt_id', '%s tt' % ref,
                                                      unique=True,
                                                      notnull=True))
            self.assertRaises(Exception, db.ttt.insert, vv='pydal')
            db.commit()
            id_i = db.tt.insert(vv='pydal')
            # first insert is OK
            db.ttt.insert(tt_id=id_i)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_i)
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Skip unicode on IMAP")
@unittest.skipIf(IS_MSSQL and not PY2, "Skip unicode on py3 and MSSQL")
class TestUnicode(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=['all'])
        db.define_table('tt', Field('vv'))
        vv = 'ἀγοραζε'
        id_i = db.tt.insert(vv=vv)
        row = db(db.tt.id == id_i).select().first()
        self.assertEqual(row.vv, vv)
        db.commit()
        drop(db.tt)
        db.close()


class TestParseDateTime(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=['all'])

        #: skip for adapters that use drivers for datetime parsing
        if db._adapter.parser.registered.get('datetime') is None:
            return

        parse = lambda v: db._adapter.parser.parse(v, 'datetime', 'datetime')

        dt = parse('2015-09-04t12:33:36.223245')
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 12)

        dt = parse('2015-09-04t12:33:36.223245Z')
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 12)

        dt = parse('2015-09-04t12:33:36.223245-2:0')
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 10)

        dt = parse('2015-09-04t12:33:36+1:0')
        self.assertEqual(dt.microsecond, 0)
        self.assertEqual(dt.hour, 13)

        dt = parse('2015-09-04t12:33:36.123')
        self.assertEqual(dt.microsecond, 123000)

        dt = parse('2015-09-04t12:33:36.00123')
        self.assertEqual(dt.microsecond, 1230)

        dt = parse('2015-09-04t12:33:36.1234567890')
        self.assertEqual(dt.microsecond, 123456)
        db.close()

@unittest.skipIf(IS_IMAP, "chained join unsupported on IMAP")
@unittest.skipIf(IS_TERADATA, "chained join unsupported on TERADATA")
class TestChainedJoinUNIQUE(unittest.TestCase):
    # 1:1 relation

    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=['all'])
        db.define_table('aa',Field('name'))
        db.define_table('bb',Field('aa','reference aa'),Field('name'))
        for k in ('x','y','z'):
            i = db.aa.insert(name=k)
            for j in ('u','v','w'):
                db.bb.insert(aa=i,name=k+j)
        db.commit()
        rows = db(db.aa).select()
        rows.join(db.bb.aa, fields=[db.bb.name], orderby=[db.bb.name])
        self.assertEqual(rows[0].bb[0].name, 'xu')
        self.assertEqual(rows[0].bb[1].name, 'xv')
        self.assertEqual(rows[0].bb[2].name, 'xw')
        self.assertEqual(rows[1].bb[0].name, 'yu')
        self.assertEqual(rows[1].bb[1].name, 'yv')
        self.assertEqual(rows[1].bb[2].name, 'yw')
        self.assertEqual(rows[2].bb[0].name, 'zu')
        self.assertEqual(rows[2].bb[1].name, 'zv')
        self.assertEqual(rows[2].bb[2].name, 'zw')

        rows = db(db.bb).select()
        rows.join(db.aa.id, fields=[db.aa.name])

        self.assertEqual(rows[0].aa.name, 'x')
        self.assertEqual(rows[1].aa.name, 'x')
        self.assertEqual(rows[2].aa.name, 'x')
        self.assertEqual(rows[3].aa.name, 'y')
        self.assertEqual(rows[4].aa.name, 'y')
        self.assertEqual(rows[5].aa.name, 'y')
        self.assertEqual(rows[6].aa.name, 'z')
        self.assertEqual(rows[7].aa.name, 'z')
        self.assertEqual(rows[8].aa.name, 'z')

        rows_json = rows.as_json()
        drop(db.bb)
        drop(db.aa)
        db.close()

class TestNullAdapter(unittest.TestCase):
    # Test that NullAdapter can define tables

    def testRun(self):
        db = DAL(None)
        db.define_table('no_table', Field('aa'))
        self.assertIsInstance(db.no_table.aa, Field)
        self.assertIsInstance(db.no_table['aa'], Field)
        db.close()



from ..adapters.couchdb import CouchDB
from .base import NoSQLDialect
from . import dialects


@dialects.register_for(CouchDB)
class CouchDBDialect(NoSQLDialect):
    def _and(self, first, second, query_env={}):
        return '(%s && %s)' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    def _or(self, first, second, query_env={}):
        return '(%s || %s)' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    def eq(self, first, second=None, query_env={}):
        if second is None:
            return '(%s == null)' % self.expand(first, query_env=query_env)
        return '(%s == %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def ne(self, first, second=None, query_env={}):
        if second is None:
            return '(%s != null)' % self.expand(first, query_env=query_env)
        return '(%s != %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def comma(self, first, second, query_env={}):
        return '%s + %s' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))



import datetime
from .._compat import integer_types, basestring
from ..adapters.base import SQLAdapter
from ..helpers.methods import use_common_filters
from ..objects import Expression, Field, Table, Select
from . import Dialect, dialects, sqltype_for

long = integer_types[-1]


class CommonDialect(Dialect):
    quote_template = '%s'

    def _force_bigints(self):
        if 'big-id' in self.types and 'reference' in self.types:
            self.types['id'] = self.types['big-id']
            self.types['reference'] = self.types['big-reference']

    def quote(self, val):
        return self.quote_template % val

    def varquote(self, val):
        return val

    def sequence_name(self, tablename):
        return self.quote('%s_sequence' % tablename)

    def trigger_name(self, tablename):
        return '%s_sequence' % tablename

    def coalesce_zero(self, val, query_env={}):
        return self.coalesce(val, [0], query_env)


@dialects.register_for(SQLAdapter)
class SQLDialect(CommonDialect):
    quote_template = '"%s"'
    true = "T"
    false = "F"
    true_exp = "1"
    false_exp = "0"
    dt_sep = " "

    @sqltype_for('string')
    def type_string(self):
        return 'VARCHAR(%(length)s)'

    @sqltype_for('boolean')
    def type_boolean(self):
        return 'CHAR(1)'

    @sqltype_for('text')
    def type_text(self):
        return 'TEXT'

    @sqltype_for('json')
    def type_json(self):
        return self.types['text']

    @sqltype_for('password')
    def type_password(self):
        return self.types['string']

    @sqltype_for('blob')
    def type_blob(self):
        return 'BLOB'

    @sqltype_for('upload')
    def type_upload(self):
        return self.types['string']

    @sqltype_for('integer')
    def type_integer(self):
        return 'INTEGER'

    @sqltype_for('bigint')
    def type_bigint(self):
        return self.types['integer']

    @sqltype_for('float')
    def type_float(self):
        return 'FLOAT'

    @sqltype_for('double')
    def type_double(self):
        return 'DOUBLE'

    @sqltype_for('decimal')
    def type_decimal(self):
        return 'NUMERIC(%(precision)s,%(scale)s)'

    @sqltype_for('date')
    def type_date(self):
        return 'DATE'

    @sqltype_for('time')
    def type_time(self):
        return 'TIME'

    @sqltype_for('datetime')
    def type_datetime(self):
        return 'TIMESTAMP'

    @sqltype_for('id')
    def type_id(self):
        return 'INTEGER PRIMARY KEY AUTOINCREMENT'

    @sqltype_for('reference')
    def type_reference(self):
        return 'INTEGER REFERENCES %(foreign_key)s ' + \
            'ON DELETE %(on_delete_action)s %(null)s %(unique)s'

    @sqltype_for('list:integer')
    def type_list_integer(self):
        return self.types['text']

    @sqltype_for('list:string')
    def type_list_string(self):
        return self.types['text']

    @sqltype_for('list:reference')
    def type_list_reference(self):
        return self.types['text']

    @sqltype_for('big-id')
    def type_big_id(self):
        return self.types['id']

    @sqltype_for('big-reference')
    def type_big_reference(self):
        return self.types['reference']

    @sqltype_for('reference FK')
    def type_reference_fk(self):
        return ', CONSTRAINT  "FK_%(constraint_name)s" FOREIGN KEY ' + \
            '(%(field_name)s) REFERENCES %(foreign_key)s ' + \
            'ON DELETE %(on_delete_action)s'

    def alias(self, original, new):
        return ('%s AS ' + self.quote_template) % (original, new)

    def insert(self, table, fields, values):
        return 'INSERT INTO %s(%s) VALUES (%s);' % (table, fields, values)

    def insert_empty(self, table):
        return 'INSERT INTO %s DEFAULT VALUES;' % table

    def where(self, query):
        return 'WHERE %s' % query

    def update(self, table, values, where=None):
        tablename = self.writing_alias(table)
        whr = ''
        if where:
            whr = ' %s' % self.where(where)
        return 'UPDATE %s SET %s%s;' % (tablename, values, whr)

    def delete(self, table, where=None):
        tablename = self.writing_alias(table)
        whr = ''
        if where:
            whr = ' %s' % self.where(where)
        return 'DELETE FROM %s%s;' % (tablename, whr)

    def select(self, fields, tables, where=None, groupby=None, having=None,
               orderby=None, limitby=None, distinct=False, for_update=False):
        dst, whr, grp, order, limit, offset, upd = '', '', '', '', '', '', ''
        if distinct is True:
            dst = ' DISTINCT'
        elif distinct:
            dst = ' DISTINCT ON (%s)' % distinct
        if where:
            whr = ' %s' % self.where(where)
        if groupby:
            grp = ' GROUP BY %s' % groupby
            if having:
                grp += ' HAVING %s' % having
        if orderby:
            order = ' ORDER BY %s' % orderby
        if limitby:
            (lmin, lmax) = limitby
            limit = ' LIMIT %i' % (lmax - lmin)
            offset = ' OFFSET %i' % lmin
        if for_update:
            upd = ' FOR UPDATE'
        return 'SELECT%s %s FROM %s%s%s%s%s%s%s;' % (
            dst, fields, tables, whr, grp, order, limit, offset, upd)

    def count(self, val, distinct=None, query_env={}):
        return ('COUNT(%s)' if not distinct else 'COUNT(DISTINCT %s)') % \
            self.expand(val, query_env=query_env)

    def join(self, val, query_env={}):
        if isinstance(val, (Table, Select)):
            val = val.query_name(query_env.get('parent_scope', []))
        elif not isinstance(val, basestring):
            val = self.expand(val, query_env=query_env)
        return 'JOIN %s' % val

    def left_join(self, val, query_env={}):
        # Left join must always have an ON clause
        if not isinstance(val, basestring):
            val = self.expand(val, query_env=query_env)
        return 'LEFT JOIN %s' % val

    def cross_join(self, val, query_env={}):
        if isinstance(val, (Table, Select)):
            val = val.query_name(query_env.get('parent_scope', []))
        elif not isinstance(val, basestring):
            val = self.expand(val, query_env=query_env)
        return 'CROSS JOIN %s' % val

    @property
    def random(self):
        return 'Random()'

    def _as(self, first, second, query_env={}):
        return '%s AS %s' % (self.expand(first, query_env=query_env), second)

    def cast(self, first, second, query_env={}):
        return 'CAST(%s)' % self._as(first, second, query_env)

    def _not(self, val, query_env={}):
        return '(NOT %s)' % self.expand(val, query_env=query_env)

    def _and(self, first, second, query_env={}):
        return '(%s AND %s)' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    def _or(self, first, second, query_env={}):
        return '(%s OR %s)' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    def belongs(self, first, second, query_env={}):
        ftype = first.type
        first = self.expand(first, query_env=query_env)
        if isinstance(second, str):
            return '(%s IN (%s))' % (first, second[:-1])
        elif isinstance(second, Select):
            if len(second._qfields) != 1:
                raise ValueError(
                    'Subquery in belongs() must select exactly 1 column')
            sub = second._compile(query_env.get('current_scope', []))[1][:-1]
            return '(%s IN (%s))' % (first, sub)
        if not second:
            return '(1=0)'
        items = ','.join(self.expand(item, ftype, query_env=query_env)
                for item in second)
        return '(%s IN (%s))' % (first, items)

    # def regexp(self, first, second):
    #     raise NotImplementedError

    def lower(self, val, query_env={}):
        return 'LOWER(%s)' % self.expand(val, query_env=query_env)

    def upper(self, first, query_env={}):
        return 'UPPER(%s)' % self.expand(first, query_env=query_env)

    def like(self, first, second, escape=None, query_env={}):
        """Case sensitive like operator"""
        if isinstance(second, Expression):
            second = self.expand(second, 'string', query_env=query_env)
        else:
            second = self.expand(second, 'string', query_env=query_env)
            if escape is None:
                escape = '\\'
                second = second.replace(escape, escape * 2)
        return "(%s LIKE %s ESCAPE '%s')" % (
            self.expand(first, query_env=query_env), second, escape)

    def ilike(self, first, second, escape=None, query_env={}):
        """Case insensitive like operator"""
        if isinstance(second, Expression):
            second = self.expand(second, 'string', query_env=query_env)
        else:
            second = self.expand(second, 'string', query_env=query_env).lower()
            if escape is None:
                escape = '\\'
                second = second.replace(escape, escape*2)
        return "(%s LIKE %s ESCAPE '%s')" % (
            self.lower(first, query_env=query_env), second, escape)

    def _like_escaper_default(self, term):
        if isinstance(term, Expression):
            return term
        term = term.replace('\\', '\\\\')
        term = term.replace(r'%', r'\%').replace('_', r'\_')
        return term

    def startswith(self, first, second, query_env={}):
        return "(%s LIKE %s ESCAPE '\\')" % (
            self.expand(first, query_env=query_env),
            self.expand(self._like_escaper_default(second)+'%', 'string',
                query_env=query_env))

    def endswith(self, first, second, query_env={}):
        return "(%s LIKE %s ESCAPE '\\')" % (
            self.expand(first, query_env=query_env),
            self.expand('%'+self._like_escaper_default(second), 'string',
                query_env=query_env))

    def replace(self, first, tup, query_env={}):
        second, third = tup
        return 'REPLACE(%s,%s,%s)' % (
            self.expand(first, 'string', query_env=query_env),
            self.expand(second, 'string', query_env=query_env),
            self.expand(third, 'string', query_env=query_env))

    def concat(self, *items, **kwargs):
        query_env = kwargs.get('query_env', {})
        tmp = (self.expand(x, 'string', query_env=query_env) for x in items)
        return '(%s)' % ' || '.join(tmp)

    def contains(self, first, second, case_sensitive=True, query_env={}):
        if first.type in ('string', 'text', 'json'):
            if isinstance(second, Expression):
                second = Expression(
                    second.db,
                    self.concat('%', Expression(
                        second.db, self.replace(second, (r'%', r'\%'),
                            query_env=query_env)), r'%'))
            else:
                second = '%'+self._like_escaper_default(str(second))+r'%'
        elif first.type.startswith('list:'):
            if isinstance(second, Expression):
                second = Expression(
                    second.db, self.concat(r'%|', Expression(
                        second.db, self.replace(Expression(
                            second.db, self.replace(
                                second, (r'%', r'\%'), query_env)),
                                ('|', '||'))), r'|%'))
            else:
                second = str(second).replace('|', '||')
                second = '%|'+self._like_escaper_default(second)+'|%'
        op = case_sensitive and self.like or self.ilike
        return op(first, second, escape='\\', query_env=query_env)

    def eq(self, first, second=None, query_env={}):
        if second is None:
            return '(%s IS NULL)' % self.expand(first, query_env=query_env)
        return '(%s = %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def ne(self, first, second=None, query_env={}):
        if second is None:
            return '(%s IS NOT NULL)' % self.expand(first, query_env=query_env)
        return '(%s <> %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def lt(self, first, second=None, query_env={}):
        if second is None:
            raise RuntimeError("Cannot compare %s < None" % first)
        return '(%s < %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def lte(self, first, second=None, query_env={}):
        if second is None:
            raise RuntimeError("Cannot compare %s <= None" % first)
        return '(%s <= %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def gt(self, first, second=None, query_env={}):
        if second is None:
            raise RuntimeError("Cannot compare %s > None" % first)
        return '(%s > %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def gte(self, first, second=None, query_env={}):
        if second is None:
            raise RuntimeError("Cannot compare %s >= None" % first)
        return '(%s >= %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def _is_numerical(self, field_type):
        return field_type in \
            ('integer', 'float', 'double', 'bigint', 'boolean') or \
            field_type.startswith('decimal')

    def add(self, first, second, query_env={}):
        if self._is_numerical(first.type) or isinstance(first.type, Field):
            return '(%s + %s)' % (
                self.expand(first, query_env=query_env),
                self.expand(second, first.type, query_env=query_env))
        else:
            return self.concat(first, second, query_env=query_env)

    def sub(self, first, second, query_env={}):
        return '(%s - %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def mul(self, first, second, query_env={}):
        return '(%s * %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def div(self, first, second, query_env={}):
        return '(%s / %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def mod(self, first, second, query_env={}):
        return '(%s %% %s)' % (
            self.expand(first, query_env=query_env),
            self.expand(second, first.type, query_env=query_env))

    def on(self, first, second, query_env={}):
        table_rname = first.query_name(query_env.get('parent_scope', []))[0]
        if use_common_filters(second):
            second = self.adapter.common_filter(second, [first])
        return ('%s ON %s') % (table_rname,
            self.expand(second, query_env=query_env))

    def invert(self, first, query_env={}):
        return '%s DESC' % self.expand(first, query_env=query_env)

    def comma(self, first, second, query_env={}):
        return '%s, %s' % (self.expand(first, query_env=query_env),
            self.expand(second, query_env=query_env))

    def extract(self, first, what, query_env={}):
        return "EXTRACT(%s FROM %s)" % (what,
            self.expand(first, query_env=query_env))

    def epoch(self, val, query_env={}):
        return self.extract(val, 'epoch', query_env)

    def length(self, val, query_env={}):
        return "LENGTH(%s)" % self.expand(val, query_env=query_env)

    def aggregate(self, first, what, query_env={}):
        return "%s(%s)" % (what, self.expand(first, query_env=query_env))

    def not_null(self, default, field_type):
        return 'NOT NULL DEFAULT %s' % \
            self.adapter.represent(default, field_type)

    @property
    def allow_null(self):
        return ''

    def coalesce(self, first, second, query_env={}):
        expressions = [self.expand(first, query_env=query_env)] + \
            [self.expand(val, first.type, query_env=query_env)
                for val in second]
        return 'COALESCE(%s)' % ','.join(expressions)

    def raw(self, val, query_env={}):
        return val

    def substring(self, field, parameters, query_env={}):
        return 'SUBSTR(%s,%s,%s)' % (
            self.expand(field, query_env=query_env), parameters[0],
            parameters[1])

    def case(self, query, true_false, query_env={}):
        _types = {bool: 'boolean', int: 'integer', float: 'double'}
        return 'CASE WHEN %s THEN %s ELSE %s END' % (
            self.expand(query, query_env=query_env),
            self.adapter.represent(
                true_false[0], _types.get(type(true_false[0]), 'string')),
            self.adapter.represent(
                true_false[1], _types.get(type(true_false[1]), 'string')))

    def primary_key(self, key):
        return 'PRIMARY KEY(%s)' % key

    def drop_table(self, table, mode):
        return ['DROP TABLE %s;' % table._rname]

    def truncate(self, table, mode=''):
        if mode:
            mode = " %s" % mode
        return ['TRUNCATE TABLE %s%s;' % (table._rname, mode)]

    def create_index(self, name, table, expressions, unique=False):
        uniq = ' UNIQUE' if unique else ''
        with self.adapter.index_expander():
            rv = 'CREATE%s INDEX %s ON %s (%s);' % (
                uniq, self.quote(name), table._rname, ','.join(
                    self.expand(field) for field in expressions))
        return rv

    def drop_index(self, name, table):
        return 'DROP INDEX %s;' % self.quote(name)

    def constraint_name(self, table, fieldname):
        return '%s_%s__constraint' % (table, fieldname)

    def concat_add(self, tablename):
        return ', ADD '

    def writing_alias(self, table):
        return table.sql_fullref


class NoSQLDialect(CommonDialect):
    @sqltype_for('string')
    def type_string(self):
        return str

    @sqltype_for('boolean')
    def type_boolean(self):
        return bool

    @sqltype_for('text')
    def type_text(self):
        return str

    @sqltype_for('json')
    def type_json(self):
        return self.types['text']

    @sqltype_for('password')
    def type_password(self):
        return self.types['string']

    @sqltype_for('blob')
    def type_blob(self):
        return self.types['text']

    @sqltype_for('upload')
    def type_upload(self):
        return self.types['string']

    @sqltype_for('integer')
    def type_integer(self):
        return long

    @sqltype_for('bigint')
    def type_bigint(self):
        return self.types['integer']

    @sqltype_for('float')
    def type_float(self):
        return float

    @sqltype_for('double')
    def type_double(self):
        return self.types['float']

    @sqltype_for('date')
    def type_date(self):
        return datetime.date

    @sqltype_for('time')
    def type_time(self):
        return datetime.time

    @sqltype_for('datetime')
    def type_datetime(self):
        return datetime.datetime

    @sqltype_for('id')
    def type_id(self):
        return long

    @sqltype_for('reference')
    def type_reference(self):
        return long

    @sqltype_for('list:integer')
    def type_list_integer(self):
        return list

    @sqltype_for('list:string')
    def type_list_string(self):
        return list

    @sqltype_for('list:reference')
    def type_list_reference(self):
        return list

    def quote(self, val):
        return val



# -*- coding: utf-8 -*-

import re

REGEX_TYPE = re.compile(r'^(?:list:)?\w+')
REGEX_DBNAME = re.compile(r'^(\w+)(:\w+)*')
REGEX_W = re.compile(r'^\w+$')
REGEX_TABLE_DOT_FIELD = re.compile(r'^(\w+)\.(\w+)$')
REGEX_TABLE_DOT_FIELD_OPTIONAL_QUOTES = r'^"?(\w+)"?\."?(\w+)"?$'
REGEX_UPLOAD_PATTERN = r'(?P<table>\w+)\.(?P<field>\w+)\.(?P<uuidkey>[\w-]+)(\.(?P<name>\w+))?\.\w+$'
REGEX_UPLOAD_CLEANUP = '[\'"\\s;]+'
REGEX_UNPACK = r'(?<!\|)\|(?!\|)'
REGEX_PYTHON_KEYWORDS = re.compile('^(False|True|and|as|assert|break|class|'
    'continue|def|del|elif|else|except|exec|finally|for|from|global|if|import|'
    'in|is|lambda|nonlocal|not|or|pass|print|raise|return|try|while|with|yield)$')
REGEX_SELECT_AS_PARSER = r'\s+AS\s+(\S+)$'
REGEX_CONST_STRING = '("[^"]*")|' "('[^']*')"
REGEX_SEARCH_PATTERN = r'^{[^.]+\.[^.]+(\.(lt|gt|le|ge|eq|ne|contains|startswith|year|month|day|hour|minute|second))?(\.not)?}$'
REGEX_SQUARE_BRACKETS = r'^.+\[.+\]$'
REGEX_UPLOAD_EXTENSION = r'\.(\w{1,5})$'
REGEX_ALPHANUMERIC = r'^[0-9a-zA-Z]\w*$'
REGEX_CREDENTIALS = r'(?<=//)[\w.-]+([:/][^@]*)?(?=@)'
REGEX_VALID_TB_FLD = re.compile(r'^[a-zA-Z]\w*\Z')
